import math
from typing import List, Optional

from rich.text import Text
from rich.panel import Panel
from rich.measure import Measurement
from rich.console import Console, ConsoleOptions, RenderResult


def nice_axis_range(minv, maxv):
    if maxv - minv <= 0:
        if maxv == minv:
            maxv += 1
        else:
            maxv, minv = minv, maxv
    exponent = int(round(math.log10(maxv - minv))) + 1

    if exponent <= 3:
        max_exponent = int(round(math.log10(math.fabs(maxv))))
        if max_exponent <= 2:
            return (minv, maxv, 0, 3 - exponent)

    shifted_maxv = math.ceil(maxv * 10 ** (exponent + 1))
    shifted_minv = math.floor(minv * 10 ** (exponent + 1))

    # make sure datarange is even
    r = shifted_maxv - shifted_minv
    shifted_maxv += r % 2

    unshifted_minv = shifted_minv * 10 ** -(exponent + 1)
    unshifted_maxv = shifted_maxv * 10 ** -(exponent + 1)
    reshifted_minv = unshifted_minv * 10**-exponent
    reshifted_maxv = unshifted_maxv * 10**-exponent

    if reshifted_maxv - reshifted_minv > 6:
        decimals = 0
    elif shifted_maxv - shifted_minv > 1:
        decimals = 1
    else:
        decimals = 2

    return (unshifted_minv, unshifted_maxv, exponent - 1, decimals)


def fit_x_labels(width, minv, maxv, exponent):
    minv = minv * 10**-exponent
    maxv = maxv * 10**-exponent
    rng = maxv - minv

    labels = [f"{minv:.1f}", f"{maxv:.1f}"]
    values = [minv, maxv]

    for i in range(4):
        nlabels = labels.copy()
        nvalues = values.copy()
        for j in range(2**i):
            v = minv + rng * (1 + 2 * j) / (2 ** (i + 1))
            nvalues.insert(1 + 2 * j, v)
            nlabels.insert(1 + 2 * j, f"{v:.1f}")

        if sum(len(k) + 2 for k in nlabels) >= width:
            break

        labels = nlabels
        values = nvalues

    line = ["─" for _ in range(width)]
    ticks = [" " for _ in range(width + len(labels[-1]))]
    for v, l in zip(values, labels):
        index = int(round((v - minv) / rng * (width - 1)))
        line[index] = "┬"
        for i, c in enumerate(l):
            ticks[index + i] = c

    return "".join(line), "".join(ticks).strip(), (len(ticks) - len(line)) // 2


def format_exponent(num):
    if num == 0:
        return "⁰"
    pre = ""
    if num < 0:
        pre = "⁻"
        num *= -1
    out = ""
    while num:
        c = num % 10
        num //= 10
        out += "⁰¹²³⁴⁵⁶⁷⁸⁹"[c]
    return pre + out[::-1]


class RichChart:
    def __init__(
        self,
        height_scale: float = 0.5,
        height: Optional[int] = None,
        width: Optional[int] = None,
        title: Optional[str] = None,
    ):
        self.plotlines = []
        self.plots = []
        self.subplots = []
        self.width = width
        self.height = height
        self.title = title
        self.plot_width = None
        self.plot_height = None
        self.draw_width = None
        self.draw_height = None
        self.height_scale = height_scale

        self.xrange = None
        self.yrange = None

    def plot(self, x: List[float], y: List[float], c: str, index: Optional[int] = None):
        if index is not None and index < len(self.plots):
            self.plots[index] = ("plot", c, x, y)
        else:
            self.plots.append(("plot", c, x, y))

        if len(x) >= 2:
            if self.xrange:
                self.xrange = (min(self.xrange[0], min(x)), max(self.xrange[1], max(x)))
                self.yrange = (min(self.yrange[0], min(y)), max(self.yrange[1], max(y)))
            else:
                self.xrange = (min(x), max(x))
                self.yrange = (min(y), max(y))

            if self.xrange[0] == self.xrange[1]:
                self.xrange = (self.xrange[0], self.xrange[0] + 1)

            if self.yrange[0] == self.yrange[1]:
                self.yrange = (self.yrange[0], self.yrange[0] + 1)

    def hist(self, x: List[float], c: str, index: Optional[int] = None):
        if index is not None and index < len(self.plots):
            self.plots[index] = ("hist", c, x)
        else:
            self.plots.append(("hist", c, x))

        if len(x) >= 2:
            if self.xrange:
                self.xrange = (min(self.xrange[0], min(x)), max(self.xrange[1], max(x)))
                self.yrange = (min(self.yrange[0], 0), max(self.yrange[1], 1))
            else:
                self.xrange = (min(x), max(x))
                self.yrange = (0, 1)

            if self.xrange[0] == self.xrange[1]:
                self.xrange = (self.xrange[0], self.xrange[0] + 1)

    def xlim(self, xlim):
        self.xrange = xlim

        if self.xrange[0] == self.xrange[1]:
            self.xrange = (self.xrange[0], self.xrange[0] + 1)

    def ylim(self, ylim):
        self.yrange = ylim

        if self.yrange[0] == self.yrange[1]:
            self.yrange = (self.yrange[0], self.yrange[0] + 1)

    def _do_binning(self):
        self.subplots = []
        clamp = lambda x, l, u: l if x < l else u if x > u else x
        bin_width = self.plot_width + 1

        for plot in self.plots:
            if plot[0] == "plot":
                bins = [0] * bin_width
                binn = [0] * bin_width

                # binning
                for xp, yp in zip(plot[2], plot[3]):
                    qx = int(
                        (xp - self.xrange[0])
                        / float(self.xrange[1] - self.xrange[0])
                        * bin_width
                    )
                    qx = clamp(qx, 0, bin_width - 1)
                    bins[qx] += (
                        (yp - self.yrange[0])
                        / float(self.yrange[1] - self.yrange[0])
                        * (self.plot_height - 1)
                    )
                    binn[qx] += 1

                # smear bins with no data
                for i in range(0, len(binn)):
                    if binn[i] == 0:
                        if i == 0:
                            bins[i] = bins[i + 1]
                            binn[i] = binn[i + 1]
                        elif i == len(binn) - 1:
                            bins[i] = bins[i - 1]
                            binn[i] = binn[i - 1]
                        else:
                            bins[i] = bins[i - 1] + bins[i + 1]
                            binn[i] = binn[i - 1] + binn[i + 1]

                bin = [
                    clamp(int(round(s / float(n))), 0, self.plot_height - 1)
                    if n > 0
                    else 0
                    for s, n in zip(bins, binn)
                ]

            elif plot[0] == "hist":
                bin = [0] * bin_width
                # binning
                for xp in plot[2]:
                    qx = int(
                        (xp - self.xrange[0])
                        / float(self.xrange[1] - self.xrange[0])
                        * bin_width
                    )
                    qx = clamp(qx, 0, bin_width - 1)
                    bin[qx] += 1

                scale = max(bin)
                bin = [
                    clamp(int(b / scale * self.plot_height), 0, self.plot_height - 1)
                    for b in bin
                ]

            self.subplots.append((bin, plot[1]))

    def _add_plot(self, subplot):
        bin, color_str = subplot
        color_str = f"[{color_str}]"
        for i, (b1, b2) in enumerate(zip(bin[:-1], bin[1:])):
            if b1 == b2:
                if i == 0:
                    self.grid[b1][i] = f"{color_str}╶[/]"
                elif i == len(bin) - 2:
                    self.grid[b1][i] = f"{color_str}╴[/]"
                else:
                    self.grid[b1][i] = f"{color_str}─[/]"
            else:
                if b1 > b2:
                    # downslope
                    self.grid[b1][i] = f"{color_str}╮[/]"
                    for j in range(b1 - 1, b2, -1):
                        self.grid[j][i] = f"{color_str}│[/]"
                    self.grid[b2][i] = f"{color_str}╰[/]"
                else:
                    # upslope
                    self.grid[b1][i] = f"{color_str}╯[/]"
                    for j in range(b1 + 1, b2):
                        self.grid[j][i] = f"{color_str}│[/]"
                    self.grid[b2][i] = f"{color_str}╭[/]"

    def _draw_plots(self):
        self.grid = [
            [" " for _ in range(self.plot_width)] for __ in range(self.plot_height)
        ]

        for i, subplot in enumerate(self.subplots):
            self._add_plot(subplot)

        self.plotlines = ["".join(self.grid[i]) + "│" for i in range(self.plot_height)]
        self.plotlines.reverse()

    def _empty_plot(self):
        return Panel(
            "No data recorded...", width=self.draw_width, height=self.draw_height
        )

    def _too_small(self):
        return Panel(
            "Too small to display", width=self.draw_width, height=self.draw_height
        )

    def __rich_console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> "RenderResult":

        # Draw sizing
        self.draw_width = (
            options.max_width - 1
            if self.width is None
            else min(options.max_width - 1, self.width)
        )
        self.plot_width = self.draw_width - 10

        self.draw_height = (
            self.height
            if self.height is not None
            else int(self.draw_width * self.height_scale / 3)
        )  # approx terminal chars thrice as high as wide
        self.draw_height -= self.draw_height % 2
        self.plot_height = self.draw_height - 3  # top bar, axis, labels

        if self.plot_height < 3:
            yield self._too_small()
            return

        if not self.plots or not self.xrange or not self.yrange:
            yield self._empty_plot()
            return

        miny, maxy, yexponent, ydecimals = nice_axis_range(*self.yrange)
        yticks = [
            f"{miny * 10 ** -yexponent + ((maxy - miny) * 10 ** -yexponent) / (self.plot_height // 2) * i:.{ydecimals}f}"
            for i in range((self.plot_height - 1) // 2 + 1)
        ]
        yticks = [a.rjust(7)[:7] for a in yticks]

        minx, maxx, xexponent, xdecimals = nice_axis_range(*self.xrange)
        xline, xticks, xtickoffset = fit_x_labels(
            self.plot_width, minx, maxx, xexponent
        )
        topline = "─" * self.plot_width
        if self.title is not None:
            r = Text.from_markup(self.title)
            segment_len = (self.plot_width - len(r)) // 2 - 1
            segment_left = "─" * segment_len + "╴"
            segment_right = "╶" + "─" * (self.plot_width - len(r) - segment_len - 2)
            topline = f"{segment_left}{self.title}{segment_right}"

        self._do_binning()
        self._draw_plots()

        ytickexp = ""
        if yexponent != 0:
            ytickexp = f"⏨{format_exponent(yexponent)}"
        ytickexp = ytickexp.rjust(7)[:7]

        xtickexp = ""
        if xexponent != 0:
            xtickexp = f"⏨{format_exponent(xexponent)}"

        yticks = sum(([f"{ytick}┤", "       │"] for ytick in yticks), [])
        yticks.pop()
        yticks.reverse()
        assert len(yticks) == len(self.plotlines)
        lines = [f"{ytick}{line}" for ytick, line in zip(yticks, self.plotlines)]

        block = (
            f"{ytickexp}╭{topline}╮\n"
            + "\n".join(lines)
            + "\n"
            + f"╰{xline}╯\n".rjust(self.draw_width)
            + f"{xtickexp} {xticks}\n".rjust(self.draw_width)
        )

        # Build text grid
        yield Text.from_markup(block)

    def __rich_measure__(
        self, console: Console, options: ConsoleOptions
    ) -> Measurement:
        return Measurement(8, options.max_width)

    def draw(self):
        return self

    def update_chart_vars(
        self, height: Optional[int] = None, width: Optional[int] = None, **kwargs
    ):
        self.height = height or self.height
        self.width = width or self.width



from dataclasses import dataclass, field
from typing import Any, Dict, List, Tuple, Set
from cyclonedds.qos import Qos
from cyclonedds.idl._typesupport.DDS.XTypes import TypeIdentifier


@dataclass
class DiscoveredType:
    type_id: TypeIdentifier
    dtype: Any
    code: str
    nested_dtypes: Dict[str, Any]
    participants: Set[str]


@dataclass
class TypeDiscoveryData:
    types: Dict[TypeIdentifier, DiscoveredType] = field(default_factory=dict)
    topic_qosses: List[Qos] = field(default_factory=list)
    writer_qosses: List[Qos] = field(default_factory=list)
    reader_qosses: List[Qos] = field(default_factory=list)

    def add_type_id(self, participant, _type: TypeIdentifier):
        if _type not in self.types:
            self.types[_type] = DiscoveredType(_type, None, "", {}, set((participant,)))
        else:
            self.types[_type].participants.add(participant)

    def split_qos(self, qosses: List[Qos]) -> Tuple[Qos, List[Qos]]:
        if not qosses:
            return Qos(), []

        if len(qosses) == 1:
            return qosses[0], []

        head, tail = qosses[0], qosses[1:]
        shared = []

        for policy in head:
            for other in tail:
                if policy not in other:
                    break
                if policy != other[policy]:
                    break
            else:
                shared.append(policy)

        shared = Qos(*shared)

        separate = []

        for policy in qosses:
            uniq = policy - shared
            if policy != shared and uniq not in separate:
                separate.append(uniq)

        if len(separate) == 1:
            separate = []

        return shared, separate



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import ctypes as ct
from typing import Union, AnyStr, Callable, Optional, Generic, Type, TypeVar, TYPE_CHECKING

from .internal import DDS, c_call, c_callable, dds_c_t
from .core import Entity, DDSException, Listener
from .qos import _CQos, Qos, LimitedScopeQos, TopicQos
from .idl import IdlStruct, IdlUnion

from cyclonedds._clayer import ddspy_topic_create


if TYPE_CHECKING:
    import cyclonedds


_S = TypeVar("_S", bound=Union[IdlStruct, IdlUnion])

class Sample(ct.Structure):
    _fields_ = [
        ('usample', ct.c_void_p),
        ('usample_size', ct.c_size_t)
    ]

_filter_fn = c_callable(ct.c_bool, [ct.POINTER(Sample), ct.c_void_p])


class Topic(Entity, Generic[_S]):
    """Representing a Topic"""

    def __init__(
            self,
            domain_participant: 'cyclonedds.domain.DomainParticipant',
            topic_name: AnyStr,
            data_type: Type[_S],
            qos: Optional[Qos] = None,
            listener: Optional[Listener] = None):
        if qos is not None:
            if isinstance(qos, LimitedScopeQos) and not isinstance(qos, TopicQos):
                raise TypeError(f"{qos} is not appropriate for a Topic")
            elif not isinstance(qos, Qos):
                raise TypeError(f"{qos} is not a valid qos object")

        if listener is not None:
            if not isinstance(listener, Listener):
                raise TypeError(f"{listener} is not a valid listener object.")

        if not hasattr(data_type, "__idl__"):
            raise TypeError(f"{data_type} is not an idl type.")

        self.data_type = data_type
        data_type.__idl__.populate()
        data_type.__idl__.fill_type_data()

        cqos = _CQos.qos_to_cqos(qos) if qos else None
        try:
            super().__init__(
                ddspy_topic_create(
                    domain_participant._ref,
                    topic_name,
                    data_type,
                    cqos,
                    listener._ref if listener else None
                ),
                listener=listener
            )
        finally:
            if cqos:
                _CQos.cqos_destroy(cqos)

        self._keepalive_entities = [self.participant]

    def get_name(self, max_size=256) -> str:
        name = (ct.c_char * max_size)()
        name_pt = ct.cast(name, ct.c_char_p)
        ret = self._get_name(self._ref, name_pt, max_size)
        if ret < 0:
            raise DDSException(ret, f"Occurred while fetching a topic name for {repr(self)}")
        return bytes(name).split(b'\0', 1)[0].decode("ASCII")

    name = property(get_name, doc="Get topic name")

    def get_type_name(self, max_size=256) -> str:
        name = (ct.c_char * max_size)()
        name_pt = ct.cast(name, ct.c_char_p)
        ret = self._get_type_name(self._ref, name_pt, max_size)
        if ret < 0:
            raise DDSException(ret, f"Occurred while fetching a topic type name for {repr(self)}")
        return bytes(name).split(b'\0', 1)[0].decode("ASCII")

    typename = property(get_type_name, doc="Get topic type name")

    def set_topic_filter(self, callable: Callable[['cyclonedds.topic', Sample], bool]):
        """Sets a filter and filter argument on a topic.

        Parameters
        ----------
        callable : filter
            The filter function used to filter topic samples.
        topic: Topic
            The topic to set the filter function.
        Sample: Sample
            The sample that needs to be checked whether to be filtered.

        Returns
        -------
        bool
            Whether this sample is filtered.
        """
        if callable is None:
            return self._set_topic_filter(self._ref, None, None)

        def call(csample, args):
            return callable(self, self.data_type.deserialize(
                ct.string_at(csample[0].usample, csample[0].usample_size)))

        self._topic_filter = _filter_fn(call)
        self._set_topic_filter(self._ref, self._topic_filter, None)

    def set_c_topic_filter(self, c_callable):
        self._c_topic_filter = c_callable
        self._set_topic_filter(self._ref, self._c_topic_filter, None)

    def get_inconsistent_topic_status(self):
        """Get INCONSISTENT_TOPIC status

        Raises
        ------
        DDSException:
            If any error code is returned by the DDS API it is converted into an exception.

        Returns
        -------
        inconsistent_topic_status:
            The class 'inconsistent_topic_status` value.
        """
        status = dds_c_t.inconsistent_topic_status()
        ret = self._get_inconsistent_topic_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the inconsistent topic status for {repr(self)}")

    


    @c_call("dds_get_name")
    def _get_name(self, topic: dds_c_t.entity, name: ct.c_char_p, size: ct.c_size_t) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_type_name")
    def _get_type_name(self, topic: dds_c_t.entity, name: ct.c_char_p, size: ct.c_size_t) -> dds_c_t.returnv:
        pass

    @c_call("dds_set_topic_filter_and_arg")
    def _set_topic_filter(self, topic: dds_c_t.entity, callback: _filter_fn, args: ct.c_void_p) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_inconsistent_topic_status")
    def _get_inconsistent_topic_status(self, topic: dds_c_t.entity, status: ct.POINTER(dds_c_t.inconsistent_topic_status)) -> dds_c_t.returnv:
        pass



"""
 * Copyright(c) 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""
import os
import platform
from pathlib import Path
from typing import List, Optional
from dataclasses import dataclass


@dataclass
class FoundCycloneResult:
    home: Path
    library_path: Path
    include_path: Path
    binary_path: Path
    ddsc_library: Path
    idlc_executable: Optional[Path]
    idlc_library: Optional[Path]
    security_libs: List[Path]


def first_or_none(alist):
    alist = list(alist)
    if alist:
        return alist[0]


def good_directory(directory: Path):
    dir = directory.resolve()
    if not dir.exists():
        return

    include_path = dir / 'include'
    bindir = dir / 'bin'

    if not include_path.exists() or not bindir.exists():
        return

    libdir = dir / 'lib'
    if not libdir.exists():
        libdir = dir / 'lib64'
        if not libdir.exists():
            return None

    if platform.system() == 'Windows':
        ddsc_library = bindir / "ddsc.dll"
    elif platform.system() == 'Darwin':
        ddsc_library = libdir / "libddsc.dylib"
    else:
        ddsc_library = libdir / "libddsc.so"

    if not ddsc_library.exists():
        return None

    idlc_executable = first_or_none(bindir.glob("idlc*"))
    idlc_library = first_or_none(libdir.glob('libcycloneddsidl*')) or first_or_none(bindir.glob("cycloneddsidl*"))
    security_libs = list(libdir.glob("*dds_security_*")) + list(bindir.glob("*dds_security_*"))

    return FoundCycloneResult(
        home=dir,
        include_path=include_path,
        library_path=libdir,
        binary_path=bindir,
        ddsc_library=ddsc_library,
        idlc_executable=idlc_executable,
        idlc_library=idlc_library,
        security_libs=security_libs
    )


def search_cyclone_pathlike(pathlike, upone=False):
    for path in pathlike.split(os.pathsep):
        if upone:
            return good_directory(Path(path) / '..')
        else:
            return good_directory(Path(path))


def find_cyclonedds() -> Optional[FoundCycloneResult]:
    if "STANDALONE_WHEELS" in os.environ:
        dir = good_directory(Path(__file__).parent.parent / "cyclonedds-build")
        if dir:
            return dir
    if "CYCLONEDDS_HOME" in os.environ:
        dir = good_directory(Path(os.environ["CYCLONEDDS_HOME"]))
        if dir:
            return dir
    if "CycloneDDS_ROOT" in os.environ:
        dir = good_directory(Path(os.environ["CycloneDDS_ROOT"]))
        if dir:
            return dir
    if "CMAKE_PREFIX_PATH" in os.environ:
        dir = search_cyclone_pathlike(os.environ["CMAKE_PREFIX_PATH"])
        if dir:
            return dir
    if "CMAKE_LIBRARY_PATH" in os.environ:
        dir = search_cyclone_pathlike(os.environ["CMAKE_LIBRARY_PATH"])
        if dir:
            return dir
    if platform.system() != "Windows" and "LIBRARY_PATH" in os.environ:
        dir = search_cyclone_pathlike(os.environ["LIBRARY_PATH"], upone=True)
        if dir:
            return dir
    if platform.system() != "Windows" and "LD_LIBRARY_PATH" in os.environ:
        dir = search_cyclone_pathlike(os.environ["LD_LIBRARY_PATH"], upone=True)
        if dir:
            return dir
    if platform.system() == "Windows" and "PATH" in os.environ:
        dir = search_cyclone_pathlike(os.environ["PATH"], upone=True)
        if dir:
            return dir




"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import uuid
import ctypes as ct
from dataclasses import dataclass
from typing import Optional, Union, TYPE_CHECKING

from .core import Entity, DDSException, Qos, ReadCondition, ViewState, InstanceState, SampleState
from .topic import Topic
from .sub import DataReader
from .internal import dds_c_t
from .qos import _CQos
from .builtin_types import DcpsParticipant, DcpsTopic, DcpsEndpoint, endpoint_constructor, participant_constructor, topic_constructor, cqos_to_qos

from cyclonedds._clayer import ddspy_read_participant, ddspy_take_participant, ddspy_read_endpoint, ddspy_take_endpoint, ddspy_read_topic, ddspy_take_topic
from cyclonedds.idl._typesupport.DDS.XTypes import TypeIdentifier


if TYPE_CHECKING:
    import cyclonedds


class BuiltinTopic(Topic):
    """Represent a built-in CycloneDDS Topic by magic reference number."""

    def __init__(self, _ref, data_type):
        self._ref = _ref
        self.data_type = data_type

    def __del__(self):
        pass


class BuiltinDataReader(DataReader):
    """
    Builtin topics have sligtly different behaviour than normal topics, so you should use this BuiltinDataReader
    instead of the normal DataReader. They are identical in the rest of their functionality.
    """

    def __init__(self,
                 subscriber_or_participant: Union['cyclonedds.sub.Subscriber', 'cyclonedds.domain.DomainParticipant'],
                 builtin_topic: 'cyclonedds.builtin.BuiltinTopic',
                 qos: Optional['cyclonedds.core.Qos'] = None,
                 listener: Optional['cyclonedds.core.Listener'] = None) -> None:
        """Initialize the BuiltinDataReader

        Parameters
        ----------
        subscriber_or_participant: cyclonedds.sub.Subscriber, cyclonedds.domain.DomainParticipant
            The subscriber to which this reader will be added. If you supply a DomainParticipant
            a subscriber will be created for you.

        builtin_topic: cyclonedds.builtin.BuiltinTopic
            Which Builtin Topic to subscribe to. This can be one of BuiltinTopicDcpsParticipant, BuiltinTopicDcpsTopic,
            BuiltinTopicDcpsPublication or BuiltinTopicDcpsSubscription. Please note that BuiltinTopicDcpsTopic will fail if
            you built CycloneDDS without Topic Discovery.
        qos: cyclonedds.core.Qos, optional = None
            Optionally supply a Qos.
        listener: cyclonedds.core.Listener = None
            Optionally supply a Listener.
        """
        self._topic = builtin_topic

        cqos = _CQos.qos_to_cqos(qos) if qos else None
        Entity.__init__(
            self,
            self._create_reader(
                subscriber_or_participant._ref,
                builtin_topic._ref,
                cqos,
                listener._ref if listener else None
            ),
            listener=listener
        )
        self._next_condition = ReadCondition(self, ViewState.Any | SampleState.NotRead | InstanceState.Any)
        if cqos:
            _CQos.cqos_destroy(cqos)
        self._make_constructors()
        self._keepalive_entities = [self.subscriber]

    def _make_constructors(self):
        if self._topic == BuiltinTopicDcpsParticipant:
            self._readfn = ddspy_read_participant
            self._takefn = ddspy_take_participant
            self._constructor = participant_constructor
        elif self._topic == BuiltinTopicDcpsTopic:
            self._readfn = ddspy_read_topic
            self._takefn = ddspy_take_topic
            self._constructor = topic_constructor
        else:
            self._readfn = ddspy_read_endpoint
            self._takefn = ddspy_take_endpoint
            self._constructor = endpoint_constructor
        self._cqos_conv = cqos_to_qos

    def read(self, N: int = 1,
             condition: Union['cyclonedds.core.ReadCondition', 'cyclonedds.core.QueryCondition'] = None):
        """Read a maximum of N samples, non-blocking. Optionally use a read/query-condition to select which samples
        you are interested in.

        Parameters
        ----------
        N: int
            The maximum number of samples to read.
        condition: cyclonedds.core.ReadCondition, cyclonedds.core.QueryCondition, optional
            Only read samples that satisfy the supplied condition.

        Raises
        ------
        DDSException
            If any error code is returned by the DDS API it is converted into an exception.
        """
        ref = condition._ref if condition else self._ref
        ret = self._readfn(ref, N, self._constructor, self._cqos_conv)

        if type(ret) == int:
            raise DDSException(ret, f"Occurred when calling read() in {repr(self)}")

        return ret

    def take(self, N: int = 1, condition=None):
        """Take a maximum of N samples, non-blocking. Optionally use a read/query-condition to select which samples
        you are interested in.

        Parameters
        ----------
        N: int
            The maximum number of samples to read.
        condition: cyclonedds.core.ReadCondition, cyclonedds.core.QueryCondition, optional
            Only take samples that satisfy the supplied condition.

        Raises
        ------
        DDSException
            If any error code is returned by the DDS API it is converted into an exception.
        """
        ref = condition._ref if condition else self._ref
        ret = self._takefn(ref, N, self._constructor, self._cqos_conv)

        if type(ret) == int:
            raise DDSException(ret, f"Occurred when calling read() in {repr(self)}")

        return ret


_pseudo_handle = 0x7fff0000
BuiltinTopicDcpsParticipant = BuiltinTopic(_pseudo_handle + 1, DcpsParticipant)
"""Built-in topic, is published to when a new participants appear on the network."""

BuiltinTopicDcpsTopic = BuiltinTopic(_pseudo_handle + 2, DcpsEndpoint)
"""
Built-in topic, is published to when a new topic appear on the network.

Make sure cyclonedds has been built with `-DENABLE_TOPIC_DISCOVERY=ON` and
`//CycloneDDS/Domain/Discovery/EnableTopicDiscoveryEndpoints` is set in the config.
"""

BuiltinTopicDcpsPublication = BuiltinTopic(_pseudo_handle + 3, DcpsEndpoint)
"""Built-in topic, is published to when a publication happens."""

BuiltinTopicDcpsSubscription = BuiltinTopic(_pseudo_handle + 4, DcpsEndpoint)
"""Built-in topic, is published to when a subscription happens."""

__all__ = [
    "DcpsParticipant", "DcpsEndpoint", "BuiltinDataReader",
    "BuiltinTopicDcpsParticipant", "BuiltinTopicDcpsTopic",
    "BuiltinTopicDcpsPublication", "BuiltinTopicDcpsSubscription"
]



import support_modules.test_classes as tc
import support_modules.test_rec_classes as trc



def test_type_objects_and_mappings():
    for _type in tc.alltypes + trc.alltypes:
        instance = _type.__idl__.get_type_info()

        if instance is None:
            continue

        instance = instance.__class__.deserialize(instance.serialize())
        assert instance == instance.__class__.deserialize(instance.serialize())
        instance = _type.__idl__.get_type_mapping()
        instance = instance.__class__.deserialize(instance.serialize())
        assert instance == instance.__class__.deserialize(instance.serialize())


def test_deduplication_of_minimal_typeids_in_typeinfo():
    container_info = tc.ContainSameTypes.__idl__.get_type_info()
    assert container_info.minimal.dependent_typeid_count + 1 == \
           container_info.complete.dependent_typeid_count



import re
import time
import signal
import platform
from typing import Any
from queue import Queue
from datetime import timedelta, datetime

import rich_click as click
from rich.pretty import Pretty
from rich.progress import Progress, BarColumn, TextColumn, TimeElapsedColumn


class TimeDeltaParamType(click.ParamType):
    name = "timedelta"

    _wordy_format = re.compile(
        r"^\s*((?P<hours>\d+)h(?:r|rs|ours|our)?)?\s*((?P<minutes>\d+)m(?:i|in|inutes|inute)?)?\s*((?P<seconds>\d+)s(?:e|ec|econds|econd)?)?\s*$"
    )
    _colon_format = re.compile(
        r"^\s*((?P<hours>\d+)\s*:)?(\s*(?P<minutes>\d+)\s*:)?\s*(?P<seconds>\d+)\s*$"
    )

    def convert(self, value, param, ctx):
        if isinstance(value, timedelta):
            return value

        try:
            return timedelta(seconds=int(value))
        except ValueError:
            pass

        for regex in [self._wordy_format, self._colon_format]:
            m = regex.match(value)
            if m:
                return timedelta(
                    hours=int(m.group("hours") or 0),
                    minutes=int(m.group("minutes") or 0),
                    seconds=int(m.group("seconds") or 0),
                )

        self.fail(f"{value} is not a valid timedelta", param, ctx)


class RateParamType(click.ParamType):
    name = "rate:inf,mnHz,nHz,nkHz,nMHz"

    def convert(self, value, param, ctx):
        if isinstance(value, int) or value == "inf":
            return value

        if value.endswith("Hz"):
            value = value[:-2]

        try:
            return float(value)
        except ValueError:
            try:
                val = float(value[:-1])
                mul = value[-1]
            except ValueError:
                self.fail(f"{value} is not a valid rate", param, ctx)
                return

        try:
            mul = {
                "m": 0.1,  # milli
                "k": 1000,  # kilo
                "M": 1000000,  # mega
            }[mul]
        except KeyError:
            self.fail(f"{value} is not a valid rate", param, ctx)
            return

        return val * mul


class SizeParamType(click.ParamType):
    name = "size:n,nk,nM"

    def convert(self, value, param, ctx):
        if isinstance(value, int):
            return value

        try:
            return int(value)
        except ValueError:
            try:
                val = int(value[:-1])
                mul = value[-1]
            except ValueError:
                self.fail(f"{value} is not a valid size", param, ctx)
                return

        try:
            mul = {
                "k": 1000,  # kilo
                "M": 1000000,  # mega
            }[mul]
        except KeyError:
            self.fail(f"{value} is not a valid size", param, ctx)
            return

        return val * mul


class LiveData:
    def __init__(self, console) -> None:
        self.console = console
        self.delivered: bool = False
        self.terminate: bool = False
        self.entities: int = 0
        self.result: Any = None
        self.printables: Queue = Queue()


def background_progress_viewer(
    runtime: timedelta, data: LiveData, suppress_view: bool = False
):
    start = datetime.now()
    end = start + runtime
    if suppress_view:
        try:
            while end > datetime.now() and not data.terminate:
                time.sleep(0.2)
            time.sleep(0.1)
        except KeyboardInterrupt:
            data.terminate = True
        return

    progress = Progress(
        TextColumn("  "),
        TimeElapsedColumn(),
        BarColumn(),
        TextColumn(
            "[blue]Entities discovered:[/] [bold purple]{task.fields[entities]}[/]"
        ),
        console=data.console,
    )

    with progress:
        try:
            task = progress.add_task("", entities=0, total=1)
            while end > datetime.now() and not data.terminate:
                time.sleep(0.2)
                progress.update(
                    task,
                    completed=((datetime.now() - start) / runtime),
                    entities=data.entities,
                )
            progress.update(task, completed=1)
        except KeyboardInterrupt:
            data.terminate = True
            return

    time.sleep(0.1)
    data.console.print()


def background_printer(data: LiveData):
    try:
        while not data.terminate:
            time.sleep(0.2)
            while data.printables:
                data.console.print(Pretty(data.printables.get()))
        time.sleep(0.1)
    except KeyboardInterrupt:
        data.terminate = True
        return



import pytest
import time

from cyclonedds.core import Statistics
from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.sub import Subscriber, DataReader
from cyclonedds.pub import Publisher, DataWriter

from support_modules.testtopics import Message


def test_create_statistics():
    dp = DomainParticipant(0)
    tp = Topic(dp, "statistics", Message)
    dw = DataWriter(dp, tp)
    stat = Statistics(dw)
    print(f"stat = {stat}")
    assert stat.data


def test_refresh_statistics():
    dp = DomainParticipant(0)
    tp = Topic(dp, "statistics", Message)
    dw = DataWriter(dp, tp)
    stat = Statistics(dw)
    assert stat.data
    time.sleep(0.5)
    stat.refresh()
    assert stat.time != 0



import pytest
import random

from cyclonedds.domain import Domain, DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.sub import Subscriber, DataReader
from cyclonedds.pub import Publisher, DataWriter
from cyclonedds.util import duration, isgoodentity
from cyclonedds.core import Qos, Policy


from support_modules.testtopics import Message, MessageKeyed

def test_reader_initialize():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    sub = Subscriber(dp)
    dr = DataReader(sub, tp)

    assert isgoodentity(dr)

def test_reader_initialize_direct():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    dr = DataReader(dp, tp)

    assert isgoodentity(dr)


def test_reader_read():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message__DONOTPUBLISH", Message)
    sub = Subscriber(dp)
    dr = DataReader(sub, tp)

    assert len(dr.read()) == 0


def test_reader_take():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message__DONOTPUBLISH", Message)
    sub = Subscriber(dp)
    dr = DataReader(sub, tp)

    assert len(dr.take()) == 0


def test_reader_waitforhistoricaldata():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message__DONOTPUBLISH", Message)
    sub = Subscriber(dp)
    dr = DataReader(sub, tp)

    assert dr.wait_for_historical_data(duration(milliseconds=5))


def test_reader_resizebuffer():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message__DONOTPUBLISH", Message)
    sub = Subscriber(dp)
    dr = DataReader(sub, tp)

    assert len(dr.read(N=100)) == 0
    assert len(dr.read(N=200)) == 0
    assert len(dr.read(N=100)) == 0


def test_reader_invalid():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message__DONOTPUBLISH", Message)
    sub = Subscriber(dp)
    dr = DataReader(sub, tp)

    with pytest.raises(TypeError):
        dr.read(-1)

    with pytest.raises(TypeError):
        dr.take(-1)


def test_reader_many_instances():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageKeyed", MessageKeyed)
    sub = Subscriber(dp)
    pub = Publisher(dp)
    dr = DataReader(sub, tp)
    dw = DataWriter(pub, tp)

    for i in range(50): # use a value which exceeds ddsc hopscotch buffer size
        msg = MessageKeyed(i, "Hello")
        dw.write(msg)
        assert dr.read_next() == msg


def test_reader_readnext_takenext():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    sub = Subscriber(dp)
    pub = Publisher(dp)
    dr = DataReader(sub, tp)
    dw = DataWriter(pub, tp)

    msg = Message("Hello")
    dw.write(msg)

    assert dr.read_next() == msg
    assert dr.read_next() is None
    dw.write(msg)
    assert dr.take_next() == msg
    assert dr.take_next() is None


def test_reader_readiter():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    sub = Subscriber(dp)
    pub = Publisher(dp)
    dr = DataReader(sub, tp)
    dw = DataWriter(pub, tp)

    msg = Message("Hello")
    dw.write(msg)

    read = False

    for msgr in dr.read_iter(timeout=duration(milliseconds=10)):
        assert not read
        assert msg == msgr
        read = True


def test_reader_takeiter():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    sub = Subscriber(dp)
    pub = Publisher(dp)
    dr = DataReader(sub, tp)
    dw = DataWriter(pub, tp)

    msg = Message("Hello")
    dw.write(msg)

    read = False

    for msgr in dr.take_iter(timeout=duration(milliseconds=10)):
        assert not read
        assert msg == msgr
        read = True


def _make_reader_without_saving_deps():
    tp = Topic(DomainParticipant(0), "Message", Message)
    return DataReader(Subscriber(tp.participant), tp)


def test_reader_keepalive_parents():
    dr = _make_reader_without_saving_deps()

    msg = Message("Hello")
    dw = DataWriter(dr.participant, dr.topic)
    dw.write(msg)

    assert dr.read_next() == msg


def test_reader_wrong_usage_errors():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    sub = Subscriber(dp)
    pub = Publisher(dp)

    with pytest.raises(TypeError):
        DataReader(False, tp)

    with pytest.raises(TypeError):
        DataReader(sub, False)

    with pytest.raises(TypeError):
        DataReader(pub, tp)

    with pytest.raises(TypeError):
        DataReader(dp, tp, qos=False)

    with pytest.raises(TypeError):
        DataReader(dp, tp, listener=False)


def test_get_matched_publications():
    dp = DomainParticipant(0)
    tp = Topic(dp, f"Message{random.randint(1000000,9999999)}", Message)
    dr = DataReader(dp, tp)

    rand_dw = random.randint(0, 20)
    dw = []
    for i in range(rand_dw):
        dw.append(DataWriter(dp, tp))

    matched = dr.get_matched_publications()
    assert len(matched) == rand_dw


def test_get_matched_publication_data():
    dp = DomainParticipant(0)
    tp = Topic(dp, f"Message{random.randint(1000000,9999999)}", Message)
    dr = DataReader(dp, tp)
    dw = DataWriter(dp, tp)

    matched_handles = dr.get_matched_publications()
    for handle in matched_handles:
        matched_data = dr.get_matched_publication_data(handle)
        assert matched_data is not None



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause


This tool is used in wheels to perform the idlc command.
The idlc binary is under cyclonedds/.libs/idlc, libs are under library_path.parent
"""

import os
import sys
import platform
import cyclonedds
from pathlib import Path


libdir = Path(cyclonedds.__file__).resolve().parent / '.libs'
idlc = (libdir / 'idlc.exe') if platform.system() == "Windows" else (libdir / 'idlc')


def command():
    if not idlc.exists():
        print("Python idlc entrypoint active but cyclonedds-python installation does not include idlc executable!")
        sys.exit(1)

    environ = os.environ.copy()

    from cyclonedds.__library__ import library_path

    if platform.system() == "Windows":
        environ["PATH"] = ";".join([str(library_path.parent)] + environ.get("PATH", "").split(";"))
    elif platform.system() == "Darwin":
        environ["DYLD_LIBRARY_PATH"] = ":".join([str(library_path.parent)] + environ.get("DYLD_LIBRARY_PATH", "").split(":"))
    else:
        environ["LD_LIBRARY_PATH"] = ":".join([str(library_path.parent)] + environ.get("LD_LIBRARY_PATH", "").split(":"))

    os.execvpe(idlc, sys.argv, environ)



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import ctypes as ct
from typing import List, Optional

from .internal import c_call, dds_c_t, dds_domain_default
from .core import Entity, DDSException, Listener
from .topic import Topic
from .qos import _CQos, LimitedScopeQos, DomainParticipantQos, Qos


class Domain(Entity):
    """A Domain represents a DDS domain with a set configuration. On the network a Domain
    is nothing more than an integer id. DDS domains are guaranteed to never mix, allowing
    logical separation of parts of your application.
    """

    def __init__(self, domain_id: int, config: Optional[str] = None):
        """Initialize a domain with domain id and configuration. The configuration is either
        a xml string or an url to a xml file.
        """
        self._id = domain_id
        if config is not None:
            super().__init__(self._create_domain(dds_c_t.domainid(domain_id), config.encode("ascii")))
        else:
            super().__init__(self._create_domain(dds_c_t.domainid(domain_id), None))

    def get_participants(self) -> List[Entity]:
        """Get all local participants of a domain."""
        num_participants = self._lookup_participant(self._id, None, 0)
        if num_participants < 0:
            raise DDSException(num_participants, f"Occurred when getting the number of participants of domain {self._id}")
        elif num_participants == 0:
            return []

        participants_list = (dds_c_t.entity * num_participants)()

        ret = self._lookup_participant(
            self._id,
            ct.cast(ct.byref(participants_list), ct.POINTER(dds_c_t.entity)),
            num_participants
        )

        if ret >= 0:
            return [Entity.get_entity(participants_list[i]) for i in range(ret)]

        raise DDSException(ret, f"Occurred when getting the participants of domain {self._id}")

    @c_call("dds_create_domain")
    def _create_domain(self, id: dds_c_t.domainid, config: ct.c_char_p) -> dds_c_t.entity:
        pass

    @c_call("dds_lookup_participant")
    def _lookup_participant(self, id: dds_c_t.domainid, participants: ct.POINTER(dds_c_t.entity), size: ct.c_size_t) \
            -> dds_c_t.entity:
        pass


class DomainParticipant(Entity):
    """The DomainParticipant is the central entrypoint for any DDS Application.
    It serves as root entity for all other entities.
    """

    def __init__(self, domain_id: int = dds_domain_default, qos: Optional[Qos] = None,
                 listener: Optional[Listener] = None):
        """Initialize a DomainParticipant.

        Parameters
        ----------
        domain_id: int, optional, default as per Cyclone DDS configuration
            The DDS Domain to use
        qos: cyclonedds.qos.Qos, optional, default None
            Apply DomainParticipant Qos.
        listener: cyclonedds.core.Listener, optional, default None
            Attach a Listener to the participant
        """
        if qos is not None:
            if isinstance(qos, LimitedScopeQos) and not isinstance(qos, DomainParticipantQos):
                raise TypeError(f"{qos} is not appropriate for a DomainParticipant")
            elif not isinstance(qos, Qos):
                raise TypeError(f"{qos} is not a valid qos object")

        if listener is not None:
            if not isinstance(listener, Listener):
                raise TypeError(f"{listener} is not a valid listener object.")

        cqos = _CQos.qos_to_cqos(qos) if qos else None
        try:
            super().__init__(self._create_participant(domain_id, cqos, listener._ref if listener else None),
                             listener=listener)
        finally:
            if cqos:
                _CQos.cqos_destroy(cqos)

    @c_call("dds_create_participant")
    def _create_participant(self, domain_id: dds_c_t.domainid, qos: dds_c_t.qos_p, listener: dds_c_t.listener_p) \
            -> dds_c_t.entity:
        pass



"""
 * Copyright(c) 2025 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import ctypes as ct
from typing import Any, Optional

from .core import DDSException
from .internal import DDS, c_call, dds_c_t, qos_kind
from .qos import (
    DataReaderQos, DataWriterQos, DomainParticipantQos, PublisherQos, Qos,
    SubscriberQos, TopicQos, _CQos,
)


class QosProvider(DDS):
    """
    Representing QosProvider
    """

    _qos_provider: dds_c_t.qos_provider_p

    def __init__(
            self,
            path: Any,
            scope: Optional[str] = ""):
        bpath = path
        if not isinstance(path, str) and not isinstance(path, bytes):
            bpath = path.__str__().encode('utf-8')
        elif isinstance(path, str):
            bpath = path.encode('utf-8')
        bscope = scope.encode('utf-8')
        self._qos_provider = dds_c_t.qos_provider_p(0)
        ret = self._create_qos_provider_scope(
            bpath, ct.byref(self._qos_provider), bscope
        )
        if ret < 0 or not self._qos_provider:
            raise DDSException(ret, f"Occured while initializing {repr(self)}")

    def _get_qos_kind(self, key: str, kind: qos_kind) -> Qos:
        bkey = key.encode('utf-8')
        cqos = dds_c_t.qos_p(0)
        ret = self._qos_provider_get_qos(
            self._qos_provider, kind, bkey, ct.byref(cqos)
        )
        if ret < 0:
            raise DDSException(ret, f"Occured trying to get_qos {repr(self)}")
        return _CQos.cqos_to_qos(cqos)

    def get_datawriter_qos(self, key: str = "") -> 'DataWriterQos':
        return self._get_qos_kind(key, qos_kind.DDS_WRITER_QOS).datawriter()

    def get_datareader_qos(self, key: str = "") -> 'DataReaderQos':
        return self._get_qos_kind(key, qos_kind.DDS_READER_QOS).datareader()

    def get_domain_participant_qos(self, key: str = "") -> 'DomainParticipantQos':
        return self._get_qos_kind(key, qos_kind.DDS_PARTICIPANT_QOS).domain_participant()

    def get_publisher_qos(self, key: str = "") -> 'PublisherQos':
        return self._get_qos_kind(key, qos_kind.DDS_PUBLISHER_QOS).publisher()

    def get_subscriber_qos(self, key: str = "") -> 'SubscriberQos':
        return self._get_qos_kind(key, qos_kind.DDS_SUBSCRIBER_QOS).subscriber()

    def get_topic_qos(self, key: str = "") -> 'TopicQos':
        return self._get_qos_kind(key, qos_kind.DDS_TOPIC_QOS).topic()

    def __del__(self) -> None:
        self._delete_qos_provider(self._qos_provider)

    @c_call("dds_create_qos_provider_scope")
    def _create_qos_provider_scope(
        self, path: ct.c_char_p, provider: ct.POINTER(dds_c_t.qos_provider_p),
        scope: ct.c_char_p
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_delete_qos_provider")
    def _delete_qos_provider(
        self, provider: dds_c_t.qos_provider_p
    ) -> None:
        pass

    @c_call("dds_qos_provider_get_qos")
    def _qos_provider_get_qos(
        self, provider: dds_c_t.qos_provider_p, qos_type: dds_c_t.qos_kind,
        key: ct.c_char_p, qos: ct.POINTER(dds_c_t.qos_p)
    ) -> dds_c_t.returnv:
        pass



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import ctypes as ct
import asyncio
import concurrent.futures
from typing import AsyncGenerator, List, Optional, TypeVar, Union, Generator, Generic, TYPE_CHECKING
import uuid

from .core import Entity, Listener, DDSException, WaitSet, ReadCondition, QueryCondition, SampleState, InstanceState, ViewState
from .domain import DomainParticipant
from .topic import Topic
from .internal import c_call, dds_c_t, InvalidSample
from .qos import _CQos, Qos, LimitedScopeQos, SubscriberQos, DataReaderQos
from .util import duration
from .builtin_types import DcpsEndpoint, endpoint_constructor, cqos_to_qos

from cyclonedds._clayer import ddspy_read, ddspy_take, ddspy_read_handle, ddspy_take_handle, ddspy_lookup_instance, ddspy_get_matched_publication_data


if TYPE_CHECKING:
    import cyclonedds


class Subscriber(Entity):
    def __init__(
            self,
            domain_participant: 'cyclonedds.domain.DomainParticipant',
            qos: Optional[Qos] = None,
            listener: Optional[Listener] = None):
        if not isinstance(domain_participant, DomainParticipant):
            raise TypeError(f"{domain_participant} is not a cyclonedds.domain.DomainParticipant.")

        if qos is not None:
            if isinstance(qos, LimitedScopeQos) and not isinstance(qos, SubscriberQos):
                raise TypeError(f"{qos} is not appropriate for a Subscriber")
            elif not isinstance(qos, Qos):
                raise TypeError(f"{qos} is not a valid qos object")

        if listener is not None:
            if not isinstance(listener, Listener):
                raise TypeError(f"{listener} is not a valid listener object.")

        cqos = _CQos.qos_to_cqos(qos) if qos else None
        try:
            super().__init__(
                self._create_subscriber(
                    domain_participant._ref,
                    cqos,
                    listener._ref if listener else None
                ),
                listener=listener
            )
        finally:
            if cqos:
                _CQos.cqos_destroy(cqos)

        self._keepalive_entities = [self.participant]

    def notify_readers(self):
        ret = self._notify_readers(self._ref)
        if ret < 0:
            raise DDSException(ret, f"Occurred while reading data in {repr(self)}")

    @c_call("dds_create_subscriber")
    def _create_subscriber(self, domain_participant: dds_c_t.entity, qos: dds_c_t.qos_p,
                           listener: dds_c_t.listener_p) -> dds_c_t.entity:
        pass

    @c_call("dds_notify_readers")
    def _notify_readers(self, subsriber: dds_c_t.entity) -> dds_c_t.returnv:
        pass


_T = TypeVar('_T')

class DataReader(Entity, Generic[_T]):
    """Subscribe to a topic and read/take the data published to it.

    All returned samples are annotated with the :class:`sample.sample_info<cyclonedds.internal.SampleInfo>` attribute.
    """

    def __init__(
            self,
            subscriber_or_participant: Union['cyclonedds.sub.Subscriber', 'cyclonedds.domain.DomainParticipant'],
            topic: Topic[_T],
            qos: Optional[Qos] = None,
            listener: Optional[Listener] = None):
        """
        Parameters
        ----------
        subscriber_or_participant: cyclonedds.sub.Subscriber, cyclonedds.domain.DomainParticipant
            The subscriber to which this reader will be added. If you supply a DomainParticipant a subscriber
            will be created for you.
        builtin_topic: cyclonedds.builtin.BuiltinTopic
            Which Builtin Topic to subscribe to. This can be one of BuiltinTopicDcpsParticipant, BuiltinTopicDcpsTopic,
            BuiltinTopicDcpsPublication or BuiltinTopicDcpsSubscription. Please note that BuiltinTopicDcpsTopic will fail if
            you built CycloneDDS without Topic Discovery.
        qos: cyclonedds.core.Qos, optional = None
            Optionally supply a Qos.
        listener: cyclonedds.core.Listener = None
            Optionally supply a Listener.
        """
        if not isinstance(subscriber_or_participant, (Subscriber, DomainParticipant)):
            raise TypeError(f"{subscriber_or_participant} is not a cyclonedds.domain.DomainParticipant"
                            " or cyclonedds.sub.Subscriber.")

        if not isinstance(topic, Topic):
            raise TypeError(f"{topic} is not a cyclonedds.topic.Topic.")

        if qos is not None:
            if isinstance(qos, LimitedScopeQos) and not isinstance(qos, DataReaderQos):
                raise TypeError(f"{qos} is not appropriate for a DataReader")
            elif not isinstance(qos, Qos):
                raise TypeError(f"{qos} is not a valid qos object")

        if listener is not None:
            if not isinstance(listener, Listener):
                raise TypeError(f"{listener} is not a valid listener object.")

        cqos = _CQos.qos_to_cqos(qos) if qos else None
        try:
            super().__init__(
                self._create_reader(
                    subscriber_or_participant._ref,
                    topic._ref,
                    cqos,
                    listener._ref if listener else None
                ),
                listener=listener
            )
        finally:
            if cqos:
                _CQos.cqos_destroy(cqos)
        self._topic = topic
        self._topic_ref = topic._ref
        self._next_condition = None
        self._keepalive_entities = [self.subscriber, topic]
        self._constructor = None

    @property
    def topic(self) -> Topic[_T]:
        return self._topic

    def read(self, N: int = 1, condition: Entity = None, instance_handle: int = None) -> List[_T]:
        """Read a maximum of N samples, non-blocking. Optionally use a read/query-condition to select which samples
        you are interested in.

        Reading samples does not remove them from the :class:`DataReader's<DataReader>` receive queue. So read
        methods may return the same sample in multiple calls.

        Parameters
        ----------
        N: int
            The maximum number of samples to read.
        condition: cyclonedds.core.ReadCondition, cyclonedds.core.QueryCondition, optional
            Only read samples that satisfy the supplied condition.

        Raises
        ------
        DDSException
            If any error code is returned by the DDS API it is converted into an exception.
        """
        use_reader = self._ref
        use_mask = SampleState.Any | ViewState.Any | InstanceState.Any
        if isinstance(condition, ReadCondition):
            use_reader = condition.reader._ref
            use_mask = condition.mask
        elif isinstance(condition, QueryCondition):
            use_reader = condition._ref
            use_mask = condition.mask

        if instance_handle is not None:
            ret = ddspy_read_handle(use_reader, use_mask, N, instance_handle)
        else:
            ret = ddspy_read(use_reader, use_mask, N)

        if type(ret) == int:
            raise DDSException(ret, f"Occurred while reading data in {repr(self)}")

        samples = []
        for (data, info) in ret:
            if info.valid_data:
                samples.append(self._topic.data_type.deserialize(data))
                samples[-1].sample_info = info
            else:
                samples.append(InvalidSample(self._topic.data_type.deserialize_key(data), info))
        return samples

    def take(self, N: int = 1, condition: Entity = None, instance_handle: int = None) -> List[_T]:
        """Take a maximum of N samples, non-blocking. Optionally use a read/query-condition to select which samples
        you are interested in.

        Taking samples removes them from the :class:`DataReader's<DataReader>` receive queue. So take methods will
        not return the same sample more than once.

        Parameters
        ----------
        N: int
            The maximum number of samples to read.
        condition: cyclonedds.core.ReadCondition, cyclonedds.core.QueryCondition, optional
            Only take samples that satisfy the supplied condition.

        Raises
        ------
        DDSException
            If any error code is returned by the DDS API it is converted into an exception.
        """
        use_reader = self._ref
        use_mask = SampleState.Any | ViewState.Any | InstanceState.Any
        if isinstance(condition, ReadCondition):
            use_reader = condition.reader._ref
            use_mask = condition.mask
        elif isinstance(condition, QueryCondition):
            use_reader = condition._ref
            use_mask = condition.mask

        if instance_handle is not None:
            ret = ddspy_take_handle(use_reader, use_mask, N, instance_handle)
        else:
            ret = ddspy_take(use_reader, use_mask, N)

        if type(ret) == int:
            raise DDSException(ret, f"Occurred while taking data in {repr(self)}")

        samples = []
        for (data, info) in ret:
            if info.valid_data:
                samples.append(self._topic.data_type.deserialize(data))
                samples[-1].sample_info = info
            else:
                samples.append(InvalidSample(self._topic.data_type.deserialize_key(data), info))
        return samples

    def read_next(self) -> Optional[_T]:
        """Shortcut method to read exactly one sample or return None.

        Raises
        ------
        DDSException
            If any error code is returned by the DDS API it is converted into an exception.
        """
        self._next_condition = self._next_condition or \
            ReadCondition(self, ViewState.Any | SampleState.NotRead | InstanceState.Alive)
        samples = self.read(condition=self._next_condition)
        if samples:
            return samples[0]
        return None

    def take_next(self) -> Optional[_T]:
        """Shortcut method to take exactly one sample or return None.

        Raises
        ------
        DDSException
            If any error code is returned by the DDS API it is converted into an exception.
        """
        self._next_condition = self._next_condition or \
            ReadCondition(self, ViewState.Any | SampleState.NotRead | InstanceState.Alive)
        samples = self.take(condition=self._next_condition)
        if samples:
            return samples[0]
        return None

    def read_iter(self, condition=None, timeout: int = None) -> Generator[_T, None, None]:
        """Shortcut method to iterate reading samples. Iteration will stop once the timeout you supply expires.
        Every time a sample is received the timeout is reset.

        Raises
        ------
        DDSException
            If any error code is returned by the DDS API it is converted into an exception.
        """
        waitset = WaitSet(self.participant)
        condition = condition or ReadCondition(self, ViewState.Any | InstanceState.Alive | SampleState.NotRead)
        waitset.attach(condition)
        timeout = timeout or duration(weeks=99999)

        while True:
            while True:
                a = self.read(condition=condition)
                if not a:
                    break
                yield a[0]
            if waitset.wait(timeout) == 0:
                break

    def read_one(self, condition=None, timeout: int = None) -> _T:
        """Shortcut method to block and take exactly one sample or raise a timeout"""
        sample = next(self.read_iter(condition=condition, timeout=timeout))
        if sample is None:
            raise TimeoutError()
        return sample

    def take_iter(self, condition=None, timeout: int = None) -> Generator[_T, None, None]:
        """Shortcut method to iterate taking samples. Iteration will stop once the timeout you supply expires.
        Every time a sample is received the timeout is reset.

        Raises
        ------
        DDSException
            If any error code is returned by the DDS API it is converted into an exception.
        """
        waitset = WaitSet(self.participant)
        condition = condition or ReadCondition(self, ViewState.Any | InstanceState.Alive | SampleState.NotRead)
        waitset.attach(condition)
        timeout = timeout or duration(weeks=99999)

        while True:
            while True:
                a = self.take(condition=condition)
                if not a:
                    break
                yield a[0]
            if waitset.wait(timeout) == 0:
                break

    def take_one(self, condition=None, timeout: int = None) -> _T:
        """Shortcut method to block and take exactly one sample or raise a timeout"""
        sample = next(self.take_iter(condition=condition, timeout=timeout))
        if sample is None:
            raise TimeoutError()
        return sample

    async def read_aiter(self, condition=None, timeout: int = None) -> AsyncGenerator[_T, None]:
        """Shortcut method to async iterate reading samples. Iteration will stop once the timeout you supply expires.
        Every time a sample is received the timeout is reset.

        Raises
        ------
        DDSException
            If any error code is returned by the DDS API it is converted into an exception.
        """
        waitset = WaitSet(self.participant)
        condition = condition or ReadCondition(self, ViewState.Any | InstanceState.Alive | SampleState.NotRead)
        waitset.attach(condition)
        timeout = timeout or duration(weeks=99999)

        loop = asyncio.get_running_loop()
        with concurrent.futures.ThreadPoolExecutor() as pool:
            while True:
                while True:
                    a = self.read(condition=condition)
                    if not a:
                        break
                    yield a[0]
                result = await loop.run_in_executor(pool, waitset.wait, timeout)
                if result == 0:
                    break

    async def take_aiter(self, condition=None, timeout: int = None) -> AsyncGenerator[_T, None]:
        """Shortcut method to async iterate taking samples. Iteration will stop once the timeout you supply expires.
        Every time a sample is received the timeout is reset.

        Raises
        ------
        DDSException
            If any error code is returned by the DDS API it is converted into an exception.
        """
        waitset = WaitSet(self.participant)
        condition = condition or ReadCondition(self, ViewState.Any | InstanceState.Alive | SampleState.NotRead)
        waitset.attach(condition)
        timeout = timeout or duration(weeks=99999)

        loop = asyncio.get_running_loop()
        with concurrent.futures.ThreadPoolExecutor() as pool:
            while True:
                while True:
                    a = self.take(condition=condition)
                    if not a:
                        break
                    yield a[0]
                result = await loop.run_in_executor(pool, waitset.wait, timeout)
                if result == 0:
                    break

    def wait_for_historical_data(self, timeout: int) -> bool:
        ret = self._wait_for_historical_data(self._ref, timeout)

        if ret == 0:
            return True
        elif ret == DDSException.DDS_RETCODE_TIMEOUT:
            return False
        raise DDSException(ret, f"Occured while waiting for historical data in {repr(self)}")

    def lookup_instance(self, sample: _T) -> Optional[int]:
        ret = ddspy_lookup_instance(self._ref, sample.serialize_key())
        if ret < 0:
            raise DDSException(ret, f"Occurred while lookup up instance from {repr(self)}")
        if ret == 0:
            return None
        return ret

    def get_matched_publications(self) -> List[int]:
        """Get instance handles of the data writers matching a reader.

        Raises
        ------
            DDSException: When the number of matching writers < 0.

        Returns
        -------
        List[int]:
            A list of instance handles of the matching data writers.
        """
        num_matched_pub = self._get_matched_publications(self._ref, None, 0)
        if num_matched_pub < 0:
            raise DDSException(num_matched_pub, f"Occurred when getting the number of matched publications of {repr(self)}")
        if num_matched_pub == 0:
            return []

        matched_pub_list = (dds_c_t.instance_handle * int(num_matched_pub))()
        matched_pub_list_pt = ct.cast(matched_pub_list, ct.POINTER(dds_c_t.instance_handle))

        ret = self._get_matched_publications(self._ref, matched_pub_list_pt, num_matched_pub)
        if ret >= 0:
            return [matched_pub_list[i] for i in range(ret)]

        raise DDSException(ret, f"Occurred when getting the matched publications of {repr(self)}")

    matched_pub = property(get_matched_publications)

    def get_matched_publication_data(self, handle) -> Optional['cyclonedds.builtin.DcpsEndpoint']:
        """Get a description of a writer matched with the provided reader.

        Parameters
        ----------
        handle: Int
            The instance handle of a writer.

        Returns
        -------
        DcpsEndpoint:
            The sample of the DcpsEndpoint built-in topic.
        """
        return ddspy_get_matched_publication_data(self._ref, handle, endpoint_constructor, cqos_to_qos)

    def get_liveliness_changed_status(self):
        """Get LIVELINESS_CHANGED status

        Raises
        ------
        DDSException

        Returns
        -------
        liveness_changed_status:
            The class 'liveness_changed_status' value.
        """
        status = dds_c_t.liveliness_changed_status()
        ret = self._get_liveliness_changed_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the liveliness changed status for {repr(self)}")

    def get_requested_deadline_missed_status(self):
        """Get REQUESTED DEALINE MISSED status

        Raises
        ------
        DDSException

        Returns
        -------
        requested_deadline_missed_status:
            The class 'requested_deadline_missed_status' value.
        """
        status = dds_c_t.requested_deadline_missed_status()
        ret = self._get_requested_deadline_missed_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the requested deadline missed status for {repr(self)}")

    def get_requested_incompatible_qos_status(self):
        """Get REQUESTED INCOMPATIBLE QOS status

        Raises
        ------
        DDSException

        Returns
        -------
        requested_incompatible_qos_status:
            The class 'requested_incompatible_qos_status' value.
        """
        status = dds_c_t.requested_incompatible_qos_status()
        ret = self._get_requested_incompatible_qos_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the requested incompatible qos status for {repr(self)}")

    def get_sample_lost_status(self):
        """Get SAMPLE LOST status

        Raises
        ------
        DDSException

        Returns
        -------
        sample_lost_status:
            The class 'sample_lost_status' value.
        """
        status = dds_c_t.sample_lost_status()
        ret = self._get_sample_lost_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the sample lost status for {repr(self)}")

    def get_sample_rejected_status(self):
        """Get SAMPLE REJECTED status

        Raises
        ------
        DDSException

        Returns
        -------
        sample_rejected_status:
            The class 'sample_rejected_status' value.
        """
        status = dds_c_t.sample_rejected_status()
        ret = self._get_sample_rejected_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the sample rejected status for {repr(self)}")

    def get_subscription_matched_status(self):
        """Get SUBSCRIPTION MATCHED status

        Raises
        ------
        DDSException

        Returns
        -------
        subscription_matched_status:
            The class 'subscription_matched_status' value.
        """
        status = dds_c_t.subscription_matched_status()
        ret = self._get_subscription_matched_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the subscription matched status for {repr(self)}")

    @c_call("dds_create_reader")
    def _create_reader(self, subscriber: dds_c_t.entity, topic: dds_c_t.entity, qos: dds_c_t.qos_p,
                       listener: dds_c_t.listener_p) -> dds_c_t.entity:
        pass

    @c_call("dds_reader_wait_for_historical_data")
    def _wait_for_historical_data(self, reader: dds_c_t.entity, max_wait: dds_c_t.duration) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_matched_publications")
    def _get_matched_publications(self, reader: dds_c_t.entity, handle: ct.POINTER(dds_c_t.instance_handle),
                                  size: ct.c_size_t) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_liveliness_changed_status")
    def _get_liveliness_changed_status(self, reader: dds_c_t.entity, status: ct.POINTER(dds_c_t.liveliness_changed_status)) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_requested_deadline_missed_status")
    def _get_requested_deadline_missed_status(self, reader: dds_c_t.entity, status: ct.POINTER(dds_c_t.requested_deadline_missed_status)) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_requested_incompatible_qos_status")
    def _get_requested_incompatible_qos_status(self, reader: dds_c_t.entity, status: ct.POINTER(dds_c_t.requested_incompatible_qos_status)) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_sample_lost_status")
    def _get_sample_lost_status(self, reader: dds_c_t.entity, status: ct.POINTER(dds_c_t.sample_lost_status)) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_sample_rejected_status")
    def _get_sample_rejected_status(self, reader: dds_c_t.entity, status: ct.POINTER(dds_c_t.sample_rejected_status)) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_subscription_matched_status")
    def _get_subscription_matched_status(self, reader: dds_c_t.entity, status: ct.POINTER(dds_c_t.subscription_matched_status)) -> dds_c_t.returnv:
        pass

__all__ = ["Subscriber", "DataReader"]



import uuid
from dataclasses import dataclass, field
from typing import List, Optional, Sequence, Set

from rich.console import Console, ConsoleOptions, RenderResult
from rich.table import Table


@dataclass
class Discoverable:
    pass


@dataclass
class PSystem:
    applications: List["PApplication"]

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        t = Table("Host", "Application", "Pid", "Participants", "Topics")
        for app in sorted(self.applications, key=lambda x: (x.hostname, x.pid)):
            r = app.row()
            if r:
                t.add_row(*r)
        yield t


@dataclass
class PApplication:
    hostname: str
    appname: str
    pid: str
    addresses: str
    participants: List["PParticipant"]

    def row(self) -> Sequence[RenderResult]:
        topics = [
            topic for topic in set().union(*(p.topics for p in self.participants))
        ]

        if not topics:
            return None

        return (
            self.hostname,
            self.appname,
            self.pid,
            "[bold magenta]" + "\n".join(str(p) for p in self.participants),
            "[bold bright_green]" + "\n".join(topics),
        )


@dataclass
class PParticipant(Discoverable):
    key: uuid.UUID
    topics: Set[str] = field(default_factory=set)
    name: Optional[str] = None

    def __str__(self):
        if self.name is not None:
            return f"{self.name} ({self.key})"
        return f"{self.key}"



import pytest
import os
from pathlib import Path
from cyclonedds.core import DDSException
from cyclonedds.qos_provider import QosProvider
from cyclonedds.qos import (
    DataReaderQos, DataWriterQos, DomainParticipantQos, PublisherQos, Qos,
    SubscriberQos, TopicQos,
)


@pytest.fixture(scope="function")
def sysdef_file(tmp_path):
    q_str = """<dds>
      <qos_library name="qoslib01">
        <qos_profile name="qosprofile01">
          <datareader_qos>
            <history><kind>KEEP_LAST_HISTORY_QOS</kind><depth>5</depth></history>
          </datareader_qos>
          <datawriter_qos>
            <history><kind>KEEP_LAST_HISTORY_QOS</kind><depth>5</depth></history>
          </datawriter_qos>
          <topic_qos>
            <resource_limits><max_instances>5</max_instances></resource_limits>
          </topic_qos>
          <publisher_qos>
            <partition><name><element>part02</element></name></partition>
          </publisher_qos>
          <subscriber_qos>
            <partition><name><element>part01</element></name></partition>
          </subscriber_qos>
          <domain_participant_qos>
            <user_data><value>ZCBzdHJpbmcgZXhhbXBsZQ==</value></user_data>
          </domain_participant_qos>
        </qos_profile>
      </qos_library>
    </dds>"""
    tmp_file = (tmp_path / "sysdef_common_test.xml")
    assert tmp_file.write_text(q_str) != 0
    return tmp_file


def test_qos_provider_create(sysdef_file):
    qp0 = QosProvider(sysdef_file, "*::qosprofile01")
    wr_q = qp0.get_datawriter_qos("qoslib01::qosprofile01")
    assert isinstance(wr_q, DataWriterQos)
    qp1 = QosProvider("<dds></dds>")
    ret = True
    try:
        wr_q = qp1.get_datawriter_qos("qoslib01::qosprofile01")
    except DDSException:
        assert ret  # nothing to get from empty definition
        ret = False
    assert not ret


def test_qos_provider_get_topic(sysdef_file):
    qp = QosProvider(sysdef_file, "*::*")
    qs = Qos.fromdict({"ResourceLimits": {"max_instances": 5}}).topic()
    assert qp.get_topic_qos("qoslib01::qosprofile01") == qs


def test_qos_provider_get_subscriber(sysdef_file):
    qp = QosProvider(str(sysdef_file), "::::")
    qs = Qos.fromdict({"Partition": {"partitions": ["part01"]}}).subscriber()
    assert qp.get_subscriber_qos("qoslib01::qosprofile01") == qs


def test_qos_provider_get_publisher(sysdef_file):
    qp = QosProvider(str(sysdef_file), "::")
    qs = Qos.fromdict({"Partition": {"partitions": ["part02"]}}).publisher()
    assert qp.get_publisher_qos("qoslib01::qosprofile01") == qs


def test_qos_provider_get_datareader(sysdef_file):
    qp = QosProvider(str(sysdef_file).encode('utf-8'), "")
    qs = Qos.fromdict({"History": {"kind": "KeepLast", "depth": 5}}).datareader()
    assert qp.get_datareader_qos("qoslib01::qosprofile01") == qs


def test_qos_provider_get_datawriter(sysdef_file):
    qp = QosProvider(sysdef_file)
    qs = Qos.fromdict({"History": {"kind": "KeepLast", "depth": 5}}).datawriter()
    assert qp.get_datawriter_qos("qoslib01::qosprofile01") == qs


def test_qos_provider_get_participant(sysdef_file):
    qp = QosProvider(sysdef_file)
    qs = Qos.fromdict({"Userdata": {"data": "ZCBzdHJpbmcgZXhhbXBsZQ=="}}).domain_participant()
    assert qp.get_domain_participant_qos("qoslib01::qosprofile01") == qs



import pytest

from cyclonedds.core import Entity, QueryCondition, SampleState, InstanceState, ViewState
from cyclonedds.util import isgoodentity

from support_modules.testtopics import Message


def test_querycondition_init(common_setup):
    qc = QueryCondition(
        common_setup.dr,
        SampleState.Any | InstanceState.Any | ViewState.Any, 
        lambda x: False
    )
    assert isgoodentity(qc)


def test_querycondition_get_mask(common_setup):
    mask = SampleState.Any | InstanceState.Any | ViewState.Any
    qc = QueryCondition(common_setup.dr, mask, lambda x: False)

    assert qc.mask == qc.get_mask() == mask

    mask = SampleState.NotRead | InstanceState.NotAliveNoWriters | ViewState.Old

    assert qc.mask == qc.get_mask() != mask

    qc = QueryCondition(common_setup.dr, mask, lambda x: False)

    assert qc.mask == qc.get_mask() == mask


def test_querycondition_get_reader(common_setup):
    qc = QueryCondition(common_setup.dr, SampleState.Any | InstanceState.Any | ViewState.Any, lambda x: False)
    assert qc.get_datareader() == common_setup.dr


# @pytest.mark.xfail(reason="TODO: implement typeless serdata")
def test_querycondition_read(common_setup):
    qc = QueryCondition(
        common_setup.dr,
        SampleState.Read | ViewState.Any | InstanceState.Any,
        lambda msg: msg.message.startswith("Goodbye")
    )

    assert not qc.triggered

    messages = [Message(message=f"Hi {i}!") for i in range(5)] + [Message(message="Goodbye")]
    for m in messages:
        common_setup.dw.write(m)

    common_setup.dr.read(N=5)
    assert not qc.triggered

    common_setup.dr.read(N=6)
    assert qc.triggered

    received = common_setup.dr.read(condition=qc)

    assert len(received) == 1 and received[0] == messages[5]



from dataclasses import dataclass
from cyclonedds.idl import IdlStruct


@dataclass
class A(IdlStruct):
    fa: int


@dataclass
class B(A):
    fb: int


def test_inheritance():
    v = B(fa=1, fb=2)
    assert v == B.deserialize(v.serialize())



from threading import Thread
import rich_click as click
from rich.console import Console
from rich.syntax import Syntax

from .utils import TimeDeltaParamType, LiveData, background_progress_viewer
from .discovery.main import type_discovery


@click.command(short_help="Fetch and display reconstructed IDL of a type over XTypes")
@click.argument("topic")
@click.option(
    "-i", "--id", "--domain-id", type=int, help="DDS Domain to inspect."
)
@click.option(
    "-r",
    "--runtime",
    type=TimeDeltaParamType(),
    default="1s",
    help="Duration of discovery scan.",
)
@click.option(
    "--suppress-progress-bar",
    type=bool,
    is_flag=True,
    help="Suppress the output of the progress bar",
)
@click.option(
    "--color",
    type=click.Choice(["auto", "standard", "256", "truecolor", "windows", "none"]),
    default="auto",
    help="""Force the command to output with/without terminal colors. By default output colours if the terminal supports it."
See the [underline blue][link=https://rich.readthedocs.io/en/stable/console.html#color-systems]Rich documentation[/link][/] for more info on what the options mean.""",
)
def typeof(topic, id, runtime, suppress_progress_bar, color):
    """Fetch and display reconstructed IDL of a type over XTypes"""
    console = Console(color_system=None if color == "none" else color)
    live = LiveData(console)

    thread = Thread(target=type_discovery, args=(live, id, runtime, topic))
    thread.start()

    background_progress_viewer(runtime, live, suppress_progress_bar)

    thread.join()

    if not live.result:
        console.print(
            "[bold red] :police_car_light: No types could be discovered over XTypes[/]"
        )
    elif len(live.result.types) > 1:
        console.print("[bold orange] :warning: Multiple type definitions exist")

    if live.result:
        for discovered_type in live.result.types.values():
            console.print(
                f"As defined in participant(s) [magenta]"
                + ", ".join(f"{p}" for p in discovered_type.participants)
            )
            console.print(Syntax(discovered_type.code, "omg-idl"))
            console.print()



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import uuid
import ctypes as ct
from dataclasses import dataclass
from typing import Optional, Union, TYPE_CHECKING

from .core import Qos
from .internal import dds_c_t
from .qos import _CQos

from cyclonedds.idl._typesupport.DDS.XTypes import TypeIdentifier

@dataclass
class DcpsParticipant:
    """
    Data sample as returned when you subscribe to the BuiltinTopicDcpsParticipant topic.

    Attributes
    ----------
    key: uuid.UUID
        Unique participant identifier
    qos: Qos
        Qos policies associated with the participant.
    """

    key: uuid.UUID
    qos: Qos


@dataclass
class DcpsTopic:
    """
    Data sample as returned when you subscribe to the BuiltinTopicDcpsTopic topic.

    Attributes
    ----------
    key:
        Unique identifier for the topic, publication or subscription endpoint.
    topic_name:
        Name of the associated topic.
    type_name:
        Name of the type.
    qos:
        Qos policies associated with the endpoint.
    typeid:
        Complete XTypes TypeIdentifier of the type, can be None.
    """

    key: uuid.UUID
    topic_name: str
    type_name: str
    qos: Qos
    type_id: Optional[TypeIdentifier]


@dataclass
class DcpsEndpoint:
    """
    Data sample as returned when you subscribe to the BuiltinTopicDcpsPublication or
    BuiltinTopicDcpsSubscription topic.

    Attributes
    ----------
    key: uuid.UUID
        Unique identifier for the topic, publication or subscription endpoint.
    participant_key: uuid.UUID
        Unique identifier of the participant the endpoint belongs to.
    participant_instance_handle: int
        Instance handle
    topic_name: str
        Name of the associated topic.
    type_name: str
        Name of the type.
    qos: Qos
        Qos policies associated with the endpoint.
    typeid: TypeIdentifier, optional
        Complete XTypes TypeIdentifier of the type, can be None.
    """

    key: uuid.UUID
    participant_key: uuid.UUID
    participant_instance_handle: int
    topic_name: str
    type_name: str
    qos: Qos
    type_id: Optional[TypeIdentifier]


def cqos_to_qos(pointer):
    p = ct.cast(pointer, dds_c_t.qos_p)
    return _CQos.cqos_to_qos(p)

def participant_constructor(keybytes, qosobject, sampleinfo):
    s = DcpsParticipant(uuid.UUID(bytes=keybytes), qos=qosobject)
    s.sample_info = sampleinfo
    return s

def endpoint_constructor(keybytes, participant_keybytes, p_instance_handle, topic_name, type_name, qosobject, sampleinfo, typeid_bytes):
    ident = None
    if typeid_bytes is not None:
        try:
            ident = TypeIdentifier.deserialize(typeid_bytes, has_header=False, use_version_2=True)
        except Exception:
            pass
        
    s = DcpsEndpoint(
        uuid.UUID(bytes=keybytes),
        uuid.UUID(bytes=participant_keybytes),
        p_instance_handle,
        topic_name,
        type_name,
        qosobject,
        ident
    )
    s.sample_info = sampleinfo
    return s

def topic_constructor(keybytes, topic_name, type_name, qosobject, sampleinfo, typeid_bytes):
    ident = None
    if typeid_bytes is not None:
        try:
            ident = TypeIdentifier.deserialize(typeid_bytes, has_header=False, use_version_2=True)
        except Exception:
            pass

    s = DcpsTopic(
        uuid.UUID(bytes=keybytes),
        topic_name,
        type_name,
        qosobject,
        ident
    )
    s.sample_info = sampleinfo
    return s



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from typing import Any, Union, Optional, Callable, List, Type, TypeVar
from . import IdlStruct, IdlUnion, IdlBitmask, IdlEnum
from ._main import IDLNamespaceScope


class AnnotationException(Exception):
    pass


T = TypeVar('T', bound=Union[Type[IdlStruct], Type[IdlUnion], Type[IdlBitmask], Type[IdlEnum]])
TIS = TypeVar('TIS', bound=Type[IdlStruct])
TIB = TypeVar('TIB', bound=Type[IdlBitmask])
TIBE = TypeVar('TIBE', Type[IdlEnum], Type[IdlBitmask])


def __annotate(cls: T, annotation: str, value: Any) -> None:
    cls.__idl_annotations__[annotation] = value


def __field_annotate(pfield: str, annotation: str, value: Any) -> None:
    if not IDLNamespaceScope.current:
        raise TypeError("Cannot annotate fields while not in class scope")

    if pfield not in IDLNamespaceScope.current["__annotations__"]:
        raise TypeError(f"Member {pfield} is not defined.")

    if pfield not in IDLNamespaceScope.current["__idl_field_annotations__"]:
        IDLNamespaceScope.current["__idl_field_annotations__"][pfield] = {}

    IDLNamespaceScope.current["__idl_field_annotations__"][pfield][annotation] = value


def default_literal(value: Any) -> Callable[[Any], None]:
    def apply_default_literal(apply_to: Any) -> None:
        __field_annotate(apply_to, "default_literal", value)
    return apply_default_literal


def key(apply_to: str) -> None:
    __field_annotate(apply_to, "key", True)


def position(apply_to: str, value: int) -> None:
    __field_annotate(apply_to, "position", value)


def member_id(apply_to: str, value: int) -> None:
    __field_annotate(apply_to, "id", value)

#Reserved Python keywords support (Issue 105):
def member_name(apply_to: str, name: str) -> None:
    __field_annotate(apply_to, "name", name)

def member_hash_id(apply_to: str, value: Optional[str] = None) -> None:
    __field_annotate(apply_to, "hash_id", value)


def external(apply_to: str) -> None:
    __field_annotate(apply_to, "external", True)


def xcdrv2(cls: T) -> T:
    __annotate(cls, "xcdrv2", True)
    return cls


def cdrv0(cls: T) -> T:
    __annotate(cls, "xcdrv2", False)
    return cls


def nested(cls: T) -> T:
    __annotate(cls, "nested", True)
    return cls


def must_understand(str: Any) -> None:
    __field_annotate(str, "must_understand", True)


def autoid(autoid_type: str) -> Callable[[T], T]:
    if autoid_type not in ["hash", "sequential"]:
        raise AnnotationException("Autoid is either 'hash' or 'sequential'.")

    def autoid_inner(cls: T) -> T:
        __annotate(cls, "autoid", autoid_type)
        return cls

    return autoid_inner


def extensibility(extensibility_type: str) -> Callable[[T], T]:
    if extensibility_type not in ["final", "mutable", "appendable"]:
        raise AnnotationException("Extensibility is either 'final', 'appendable' or 'mutable'.")

    def extensibility_inner(cls: T) -> T:
        __annotate(cls, "extensibility", extensibility_type)
        return cls

    return extensibility_inner


def final(cls: T) -> T:
    return extensibility("final")(cls)


def appendable(cls: T) -> T:
    return extensibility("appendable")(cls)


def mutable(cls: T) -> T:
    return extensibility("mutable")(cls)


def keylist(list_of_keys: List[str]) -> Callable[[TIS], TIS]:
    def keylist_inner(cls: TIS) -> TIS:
        __annotate(cls, "keylist", list_of_keys)
        return cls

    return keylist_inner


def bit_bound(amount: int) -> Callable[[TIBE], TIBE]:
    if amount <= 0 or amount > 64:
        raise AnnotationException(f"{amount} is not a valid bit_bound, must be between 1 and 64 inclusive.")

    def bit_bound_inner(cls: TIBE) -> TIBE:
        __annotate(cls, "bit_bound", amount)

        if hasattr(cls, "__idl_highest_position__") and amount <= cls.__idl_highest_position__:
            raise AnnotationException(f"Position {cls.__idl_highest_position__} is outside of the bit boundary [0, {amount}).")
        return cls

    return bit_bound_inner


__all__ = [
    "default_literal", "key", "position", "member_id", "member_hash_id", "xcdrv2", "cdrv0",
    "nested", "must_understand", "autoid", "extensibility", "final", "appendable", "mutable",
    "keylist", "bit_bound"
]



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import asyncio
import concurrent.futures
from typing import Union, Dict, Tuple
from collections import deque

from . import _clayer as cl
from .internal import feature_type_discovery
from .core import DDSException
from .domain import DomainParticipant
from .idl import IdlBitmask, IdlEnum, IdlUnion, IdlStruct
from .idl._xt_builder import XTInterpreter, XTTypeIdScanner
from .idl._typesupport.DDS.XTypes._ddsi_xt_type_object import TypeIdentifier, TypeObject


def get_types_for_typeid(participant: DomainParticipant, type_id: TypeIdentifier, timeout: int) -> \
        Tuple[Union[IdlUnion, IdlStruct], Dict[str, Union[IdlUnion, IdlStruct, IdlEnum, IdlBitmask]]]:
    """Attempt to gather the Python types that match a TypeIdentifier. This might involve several
    network roundtrips to request necessary type information.

    Returns
    -------
    (Type, Dict[str, Type])
        Returns a type that can be used in Topic creation (IdlUnion or IdlStruct subclass), as well as all
        types by name that appear at some nesting inside the main type (IdlUnion, IdlStruct, IdlEnum or IdlBitmask)

    Raises
    ------
    DDSException
        If ENABLE_TYPE_DISCOVERY is not set upon compiling Cyclone DDS this does not work.
    """
    if not feature_type_discovery:
        raise DDSException(DDSException.DDS_RETCODE_ILLEGAL_OPERATION, "CycloneDDS was not compiled with type support")

    typemap = {}
    to_resolve = deque((type_id,))

    while to_resolve:
        tid = to_resolve.pop()
        if tid in typemap:
            continue

        # ddspy_get_typeobj releases gil
        ret = cl.ddspy_get_typeobj(participant._ref, tid.serialize(use_version_2=True)[4:], timeout)
        if type(ret) == int:
            raise DDSException(ret, f"Could not fetch typeobject for {tid}")

        try:
            type_object = TypeObject.deserialize(ret, has_header=False, use_version_2=True)
        except Exception as e:
            raise DDSException(DDSException.DDS_RETCODE_ERROR, "Got invalid TypeObject from C layer.") from e

        for dep_tid in XTTypeIdScanner.find_all_typeids(type_object):
            to_resolve.append(dep_tid)

        typemap[tid] = type_object

    return XTInterpreter.xt_to_class(type_id, typemap)


async def async_get_types_for_typeid(participant: DomainParticipant, type_id: TypeIdentifier, timeout: int) -> \
        Tuple[Union[IdlUnion, IdlStruct], Dict[str, Union[IdlUnion, IdlStruct, IdlEnum, IdlBitmask]]]:
    """Async version of get_types_for_typeid. Runs in separate thread."""
    if not feature_type_discovery:
        raise DDSException(DDSException.DDS_RETCODE_ILLEGAL_OPERATION, "CycloneDDS was not compiled with type support")

    loop = asyncio.get_running_loop()
    with concurrent.futures.ThreadPoolExecutor() as pool:
        return await loop.run_in_executor(pool, get_types_for_typeid, participant, type_id, timeout)



import uuid
from cyclonedds.builtin import DcpsEndpoint, DcpsParticipant
from cyclonedds.core import Qos, Policy
from cyclonedds.idl._typesupport.DDS.XTypes import TypeIdentifier, EK_COMPLETE
from dataclasses import dataclass
from typing import List, Optional

from rich.console import Console, ConsoleOptions, RenderResult, Group, NewLine, group
from rich.panel import Panel
from rich.pretty import Pretty
from rich.padding import Padding
from rich.columns import Columns
from rich.table import Table, Column

from ..qosformat import rich_qos


def fmt_ident(tid: TypeIdentifier) -> str:
    if tid is None:
        return f"unset"

    if tid.discriminator == EK_COMPLETE:
        return f"COMPLETE {tid.equivalence_hash.hex().upper()}"
    else:
        return f"MINIMAL {tid.equivalence_hash.hex().upper()}"


@dataclass
class Discoverable:
    pass


@dataclass
class DParticipant(Discoverable):
    sample: DcpsParticipant
    topics: List["DTopic"]
    show_qos: bool
    is_self: bool = False

    def name(self):
        if Policy.EntityName in self.sample.qos:
            return self.sample.qos[Policy.EntityName].name
        return None

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        name = self.name()
        title = (
            f"[bold magenta]{name}[/] ([bold magenta]{self.sample.key}[/])"
            if name
            else f"[bold magenta]{self.sample.key}[/]"
        )
        yield Panel.fit(
            self.get_subcontent(console, options), title=f"Participant {title}"
        )

    @group()
    def get_subcontent(self, console: Console, options: ConsoleOptions) -> Group:
        yield ""

        qos = self.sample.qos
        name = None
        if Policy.EntityName in qos:
            name = qos[Policy.EntityName]
            qos = qos - Qos(name)
            name = name.name

        if self.show_qos and qos.policies:
            yield Panel.fit(
                rich_qos(qos), border_style="cyan", title="[bold bright_cyan] QoS"
            )
            yield ""

        for topic in self.topics:
            yield from topic.__rich_console__(console, options)


@dataclass
class DPubSub:
    name: Optional[str]
    endpoint: DcpsEndpoint
    qos: Qos  # without EntityName


@dataclass
class DTopic(Discoverable):
    name: str
    qos: Qos
    show_qos: bool
    publications: List[DPubSub]
    subscriptions: List[DPubSub]

    def shared_qos(self, entities: List[DPubSub]) -> Qos:
        if not entities:
            return Qos()

        if len(entities) == 1:
            return Qos(*[p for p in entities[0].qos if p not in self.qos])

        head, tail = entities[0], entities[1:]
        shared = []

        for policy in head.qos:
            if policy in self.qos:
                continue

            for other in tail:
                if policy not in other.qos:
                    break
            else:
                shared.append(policy)

        return Qos(*shared)

    def unshared_qos(self, shared_qos: Qos, entities: List[DPubSub]) -> List[Qos]:
        return [
            Qos(*[p for p in h.qos if p not in shared_qos and p not in self.qos])
            for h in entities
        ]

    @group()
    def render(self):
        pubsub = self.subscriptions + self.publications

        if not pubsub:
            return

        if self.show_qos:
            common_qos = self.shared_qos(pubsub)
            common_writer_qos = (
                self.shared_qos(self.publications) if self.publications else common_qos
            )
            common_reader_qos = (
                self.shared_qos(self.subscriptions)
                if self.subscriptions
                else common_qos
            )
            writer_qos = self.unshared_qos(common_writer_qos, self.publications)
            reader_qos = self.unshared_qos(common_reader_qos, self.subscriptions)
            common_writer_qos = common_writer_qos - common_qos
            common_reader_qos = common_reader_qos - common_qos
            writer_qos_consistent = all(not q.policies for q in writer_qos)
            reader_qos_consistent = all(not q.policies for q in reader_qos)

        typename = pubsub[0].endpoint.type_name
        if not all(t.endpoint.type_name == typename for t in pubsub):
            typename = ":police_car_light: [bold red]Inconsistent, see below[/]"
            typename_consistent = False
        else:
            typename = f"[bold magenta]{typename}[/]"
            typename_consistent = True

        type_id = pubsub[0].endpoint.type_id
        if not all(t.endpoint.type_id == type_id for t in pubsub):
            type_id = ":police_car_light: [bold orange]Inconsistent, see below[/]"
            type_id_consistent = False
        else:
            type_id = f"[bold magenta]{fmt_ident(type_id)}[/]"
            type_id_consistent = True

        properties = Table(show_header=False)
        properties.add_column("Key", justify="left", no_wrap=True)
        properties.add_column("Value", justify="right", no_wrap=True)
        properties.add_row("Typename", typename)
        properties.add_row("XTypes Type ID", type_id)

        if self.show_qos and not common_writer_qos and not common_reader_qos:
            qos_display = Table.grid()
            qos_display.add_column()
            qos_display.add_column()
            qos_display.add_row(
                Panel.fit(
                    rich_qos(self.qos),
                    border_style="cyan",
                    title="[bold bright_cyan]Topic QoS[/]",
                ),
                Panel.fit(
                    rich_qos(common_qos),
                    border_style="cyan",
                    title="[bold bright_cyan]Common Endpoint QoS[/]",
                ),
            )
        elif self.show_qos:
            qos_display = Table.grid()
            qos_display.add_column()
            qos_display.add_column()
            qos_display.add_column()
            qos_display.add_column()
            qos_display.add_row(
                Panel.fit(
                    rich_qos(self.qos),
                    border_style="cyan",
                    title="[bold bright_cyan]Topic QoS[/]",
                ),
                Panel.fit(
                    rich_qos(common_qos),
                    border_style="cyan",
                    title="[bold bright_cyan]Common Endpoint QoS[/]",
                ),
                Panel.fit(
                    rich_qos(common_writer_qos),
                    border_style="cyan",
                    title="[bold bright_cyan]Common Writer QoS[/]",
                ),
                Panel.fit(
                    rich_qos(common_reader_qos),
                    border_style="cyan",
                    title="[bold bright_cyan]Common Reader QoS[/]",
                ),
            )

        pub_display = None
        sub_display = None

        if self.publications:
            pub_display = Table(title="Writers")
            pub_display.add_column("GUID", no_wrap=True)

            if not typename_consistent or not type_id_consistent:
                pub_display.add_column("Properties")

            if self.show_qos and not writer_qos_consistent:
                pub_display.add_column("QoS")

            for i, pub in enumerate(self.publications):

                if pub.name:
                    row = [f"[bold blue]{pub.name} ({pub.endpoint.key})[/]"]
                else:
                    row = [f"[bold blue]{pub.endpoint.key}[/]"]

                if not typename_consistent or not type_id_consistent:
                    props = []

                    if pub.endpoint.type_name != typename:
                        props.append(
                            f"Typename: [bold magenta]{pub.endpoint.type_name}[/]"
                        )

                    if pub.endpoint.type_id != type_id:
                        props.append(
                            f"XTypes Type ID: [bold magenta]{fmt_ident(pub.endpoint.type_id)}[/]"
                        )

                    if not props:
                        row.append("")
                    elif len(props) == 1:
                        row.append(props[0])
                    else:
                        row.append(Group(*props))

                if (
                    self.show_qos
                    and not writer_qos_consistent
                    and writer_qos[i].policies
                ):
                    row.append(rich_qos(writer_qos[i]))

                pub_display.add_row(*row)

        if self.subscriptions:
            sub_display = Table(title="Readers")
            sub_display.add_column("GUID", no_wrap=True)

            if not typename_consistent or not type_id_consistent:
                sub_display.add_column("Properties")

            if self.show_qos and not reader_qos_consistent:
                sub_display.add_column("QoS")

            for i, sub in enumerate(self.subscriptions):
                if sub.name:
                    row = [f"[bold blue]{sub.name} ({sub.endpoint.key})[/]"]
                else:
                    row = [f"[bold blue]{sub.endpoint.key}[/]"]

                if not typename_consistent or not type_id_consistent:
                    props = []

                    if sub.endpoint.type_name != typename:
                        props.append(
                            f"Typename: [bold magenta]{sub.endpoint.type_name}[/]"
                        )

                    if sub.endpoint.type_id != type_id:
                        props.append(
                            f"XTypes Type ID: [bold magenta]{fmt_ident(sub.endpoint.type_id)}[/]"
                        )

                    if not props:
                        row.append("")
                    elif len(props) == 1:
                        row.append(props[0])
                    else:
                        row.append(Group(*props))

                if (
                    self.show_qos
                    and not reader_qos_consistent
                    and reader_qos[i].policies
                ):
                    row.append(rich_qos(reader_qos[i]))

                sub_display.add_row(*row)

        # render
        if self.show_qos:
            yield qos_display

        yield properties

        if pub_display:
            yield pub_display

        if sub_display:
            yield sub_display

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Panel.fit(self.render(), title=f"[bold green]{self.name}[/]")



from threading import Thread
import rich_click as click
from rich.console import Console

from .utils import TimeDeltaParamType, LiveData, background_progress_viewer
from .discovery.main import ls_discovery


@click.command(short_help="Scan and display DDS entities in your network")
@click.option(
    "-i", "--id", "--domain-id", type=int, help="DDS Domain to inspect."
)
@click.option(
    "-r",
    "--runtime",
    type=TimeDeltaParamType(),
    default="1s",
    help="Duration of discovery scan.",
)
@click.option(
    "-t",
    "--topic",
    type=str,
    help="Filter which entity types to display by topic name (supports regex)",
    default=".*",
)
@click.option(
    "--show-self",
    type=bool,
    is_flag=True,
    help="Show the tools own participant and subscriptions.",
)
@click.option(
    "--suppress-progress-bar",
    type=bool,
    is_flag=True,
    help="Suppress the output of the progress bar",
)
@click.option(
    "--color",
    type=click.Choice(["auto", "standard", "256", "truecolor", "windows", "none"]),
    default="auto",
    help="""Force the command to output with/without terminal colors. By default output colours if the terminal supports it."
See the [underline blue][link=https://rich.readthedocs.io/en/stable/console.html#color-systems]Rich documentation[/link][/] for more info on what the options mean.""",
)
@click.option(
    "-q",
    "--qos",
    type=bool,
    is_flag=True,
    help="Show the QoS settings of all entities.",
    default=False,
)
def ls(
    topic, id, runtime, show_self: bool, suppress_progress_bar: bool, color, qos: bool
):
    """Scan and display DDS entities in your network."""
    console = Console(color_system=None if color == "none" else color)
    live = LiveData(console)

    thread = Thread(target=ls_discovery, args=(live, id, runtime, topic, qos))
    thread.start()

    background_progress_viewer(runtime, live, suppress_progress_bar)

    thread.join()

    if live.result:
        for p in live.result:
            if p.is_self and not show_self:
                continue

            console.print(p)
            console.print()



import pytest

from cyclonedds.core import Entity
from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.util import isgoodentity
from cyclonedds.pub import DataWriter
from cyclonedds.sub import DataReader
from cyclonedds.qos import Policy, Qos
from dataclasses import dataclass

from support_modules.testtopics import Message


def test_create_topic():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)

    assert isgoodentity(tp)


def test_get_name():
    dp = DomainParticipant(0)
    tp = Topic(dp, 'MessageTopic', Message)

    assert tp.name == tp.get_name() == 'MessageTopic'


def test_get_type_name():
    dp = DomainParticipant(0)
    tp = Topic(dp, 'MessageTopic', Message)

    assert tp.typename == tp.get_type_name() == 'Message'


def filter(topic: Topic, sample: Message) -> bool:
    if "Filter" in sample.message:
        return False
    return True


def test_topic_filter():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageTopic", Message, qos=Qos(Policy.History.KeepLast(5)))
    tp.set_topic_filter(filter)
    dw = DataWriter(dp, tp)
    dr = DataReader(dp, tp)

    dw.write(Message("Nice Message"))
    dw.write(Message("Test Filtering"))
    dw.write(Message("Hello"))
    dw.write(Message("lower case filter"))

    data = str(dr.read(5))
    assert "Filter" not in data
    assert "filter" and "Hello" and "Nice" in data



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from typing import Optional
import warnings

from .entity_qos import EntityQosMapper

from cyclonedds.core import Qos


def warning_msg(message, category, filename, lineno, file=None, line=None):
    return "%s: %s\n" % (category.__name__, message)


warnings.formatwarning = warning_msg


class InapplicableQosWarning(UserWarning):
    pass


class QosPerEntity:
    def __init__(self, entity):
        self.topic_qos: Optional[Qos] = None
        self.subscriber_qos: Optional[Qos] = None
        self.publisher_qos: Optional[Qos] = None
        self.datareader_qos: Optional[Qos] = None
        self.datawriter_qos: Optional[Qos] = None
        self.entity = entity  # Keep track of the original entity for qos-all

    def entity_qos(self, qos, entity):
        if entity == "topic":
            self.topic_qos = self.check_entity_qos("topic", EntityQosMapper.topic, qos)
        elif entity == "publisher":
            self.publisher_qos = self.check_entity_qos("publisher", EntityQosMapper.pubsub, qos)
        elif entity == "subscriber":
            self.subscriber_qos = self.check_entity_qos("subscriber", EntityQosMapper.pubsub, qos)
        elif entity == "datawriter":
            self.datawriter_qos = self.check_entity_qos("datawriter", EntityQosMapper.writer, qos)
        elif entity == "datareader":
            self.datareader_qos = self.check_entity_qos("datareader", EntityQosMapper.reader, qos)
        else:  # qos-all
            for e in ["topic", "publisher", "subscriber", "datawriter", "datareader"]:
                self.entity_qos(qos, e)

    def check_entity_qos(self, entity, eqos_mapper, qos):
        eq = []
        for q in qos:
            policy_scope = f"Policy.{q.__scope__}"

            if policy_scope in eqos_mapper:
                eq.append(q)
            elif self.entity != "all" and self.entity is not None:
                warnings.warn(f"The {q} is not applicable for {entity}, will be ignored.", InapplicableQosWarning)
        return Qos(*eq)



import pytest

from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.sub import DataReader
from cyclonedds.pub import DataWriter
from cyclonedds.qos import Qos, Policy
from cyclonedds.core import DDSException


from support_modules.testtopics import XMessage, Message


def test_data_representation_writer_v0_match():
    qos = Qos(Policy.DataRepresentation(use_cdrv0_representation=True))

    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    dr = DataReader(dp, tp, qos=qos)
    dw = DataWriter(dp, tp, qos=qos)

    assert dw._use_version_2 == False

    msg = Message("Hello")
    dw.write(msg)
    assert dr.read_next() == msg


def test_data_representation_writer_v0_notdefault():
    qos = Qos(Policy.DataRepresentation(use_cdrv0_representation=True))

    dp = DomainParticipant(0)
    tp = Topic(dp, "XMessage", XMessage)
    dr = DataReader(dp, tp, qos=qos)
    dw = DataWriter(dp, tp, qos=qos)

    assert dw._use_version_2 == False

    msg = XMessage("Hello")
    dw.write(msg)
    assert dr.read_next() == msg


def test_data_representation_writer_v2_match():
    qos = Qos(Policy.DataRepresentation(use_xcdrv2_representation=True))

    dp = DomainParticipant(0)
    tp = Topic(dp, "XMessage", XMessage)
    dr = DataReader(dp, tp, qos=qos)
    dw = DataWriter(dp, tp, qos=qos)

    assert dw._use_version_2 == True

    msg = XMessage("Hello")
    dw.write(msg)
    assert dr.read_next() == msg


def test_data_representation_writer_v0_v2_unmatch():
    qosv0 = Qos(Policy.DataRepresentation(use_cdrv0_representation=True))
    qosv2 = Qos(Policy.DataRepresentation(use_xcdrv2_representation=True))

    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    dr = DataReader(dp, tp, qos=qosv0)
    dw = DataWriter(dp, tp, qos=qosv2)

    msg = Message("Hello")
    dw.write(msg)
    assert dr.read_next() == None


def test_data_representation_writer_v2_v0_unmatch():
    qosv0 = Qos(Policy.DataRepresentation(use_cdrv0_representation=True))
    qosv2 = Qos(Policy.DataRepresentation(use_xcdrv2_representation=True))

    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    dr = DataReader(dp, tp, qos=qosv2)
    dw = DataWriter(dp, tp, qos=qosv0)

    msg = Message("Hello")
    dw.write(msg)
    assert dr.read_next() == None


def test_data_representation_writer_dualreader_match():
    qosv0 = Qos(Policy.DataRepresentation(use_cdrv0_representation=True))
    qosv2 = Qos(Policy.DataRepresentation(use_xcdrv2_representation=True))
    qosv0v2 = Qos(Policy.DataRepresentation(use_cdrv0_representation=True, use_xcdrv2_representation=True))

    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    dr = DataReader(dp, tp, qos=qosv0v2)
    dwv0 = DataWriter(dp, tp, qos=qosv0)
    dwv2 = DataWriter(dp, tp, qos=qosv2)

    msg1 = Message("Hello")
    dwv0.write(msg1)
    assert dr.read_next() == msg1

    msg2 = Message("Hi!")
    dwv2.write(msg2)
    assert dr.read_next() == msg2




import pytest
import asyncio
import shutil
import concurrent
import subprocess
import json
import time
import sys
import tempfile

from cyclonedds.core import Qos, Policy


if sys.platform.startswith("win"):
    pytest.skip("DDSLS is unstable on windows", allow_module_level=True)


# Helper functions


message = "test\n 420\n [4,2,0]\n ['test','str','array','data','struct']\n [-1,183]\n ['test','string','sequence']"


def run_pubsub(args, text=message, timeout=10):
    tmp_dir = tempfile.mkdtemp()
    process = subprocess.Popen(
        [sys.executable, "-m", "cyclonedds.tools.pubsub"] + args,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=tmp_dir
    )

    try:
        if text:
            process.stdin.write(text.encode())
        stdout, stderr = process.communicate(timeout=timeout)
    except subprocess.TimeoutExpired as e:
        try:
            process.kill()
            shutil.rmtree(tmp_dir)
        except:
            # Observant readers will note that if the process is not killed
            # then we do not delete the temporary dir. This has a good reason:
            # When the process is stuck in an unkillable state and we try to remove
            # the directory, we will enter an infinite loop with permission errors on
            # windows.
            pass
        try:
            print("Pubsub command timeout")
            print("   attempting to pull from stdout:")
            process.stdout.close()
            print(process.stdout.read().decode().replace("\r", ""))
            print("   attempting to pull from stderr")
            process.stderr.close()
            print(process.stderr.read().decode().replace("\r", ""))
        except:
            pass

        raise e

    shutil.rmtree(tmp_dir)
    return {
        "stdout": stdout.decode().replace("\r", ""),
        "stderr": stderr.decode().replace("\r", ""),
        "status": process.returncode
    }


def run_ddsls(args, timeout=10):
    tmp_dir = tempfile.mkdtemp()
    process = subprocess.Popen([sys.executable, "-m", "cyclonedds.tools.ddsls"] + args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            cwd=tmp_dir
                            )
    try:
        stdout, stderr = process.communicate(timeout=timeout)
    except subprocess.TimeoutExpired as e:
        try:
            process.kill()
            shutil.rmtree(tmp_dir)
        except:
            # Observant readers will note that if the process is not killed
            # then we do not delete the temporary dir. This has a good reason:
            # When the process is stuck in an unkillable state and we try to remove
            # the directory, we will enter an infinite loop with permission errors on
            # windows.
            pass
        try:
            print("DDSLS command timeout")
            print("   attempting to pull from stdout:")
            process.stdout.close()
            print(process.stdout.read().decode().replace("\r", ""))
            print("   attempting to pull from stderr")
            process.stderr.close()
            print(process.stderr.read().decode().replace("\r", ""))
        except:
            pass
        raise e

    shutil.rmtree(tmp_dir)

    return {
        "stdout": stdout.decode().replace("\r", ""),
        "stderr": stderr.decode().replace("\r", ""),
        "status": process.returncode
    }


async def run_pubsub_ddsls_async(pubsub_args, ddsls_args, runtime):
    loop = asyncio.get_event_loop_policy().get_event_loop()
    with concurrent.futures.ThreadPoolExecutor() as pool:
        pubsub_task = loop.run_in_executor(pool, run_pubsub, ["--runtime", str(runtime-1.5)] + pubsub_args)
        ddsls_task = loop.run_in_executor(pool, run_ddsls, ["--watch", "--runtime", str(runtime)] + ddsls_args)
        await asyncio.sleep(0.3)
        return (await pubsub_task), (await ddsls_task)


def run_pubsub_ddsls(pubsub_args, ddsls_args, runtime=5):
    loop = asyncio.get_event_loop_policy().get_event_loop()
    result = loop.run_until_complete(run_pubsub_ddsls_async(pubsub_args, ddsls_args, runtime))
    return result


# tests


def test_pubsub_empty():
    pubsub = run_pubsub(["-T", "test", "--runtime", "0.5"], text=None)
    assert pubsub["stdout"] == ""
    assert pubsub["status"] == 0


def test_pubsub_topics():
    pubsub = run_pubsub(["-T", "test", "--runtime", "0.5"])

    assert "String(seq=0, keyval='test')" in pubsub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert "IntArray(seq=2, keyval=[4, 2, 0])" in pubsub["stdout"]
    assert "StrArray(seq=3, keyval=['test', 'str', 'array', 'data', 'struct'])" in pubsub["stdout"]
    assert "IntSequence(seq=4, keyval=[-1, 183])" in pubsub["stdout"]
    assert "StrSequence(seq=5, keyval=['test', 'string', 'sequence'])" in pubsub["stdout"]
    assert pubsub["status"] == 0


def test_parse_qos():
    tests = \
    [
        (
            [
                "Reliability.Reliable 10000000",
                "Durability.TransientLocal",
                "History.KeepLast 10",
                "ResourceLimits 100 -1 100",
                "PresentationAccessScope.Topic True False",
                "Lifespan 1000000",
                "Deadline seconds=1",
                "LatencyBudget 10000000",
                "Ownership.Exclusive",
                "OwnershipStrength 20",
                "Liveliness.ManualByParticipant 100000",
                "TimeBasedFilter 100000",
                "Partition Hello, world",
                "TransportPriority 1",
                "DestinationOrder.BySourceTimestamp",
                "WriterDataLifecycle False",
                "ReaderDataLifecycle 10000000 10000000",
                "DurabilityService 100000 History.KeepLast 100, 2000, 1000, 1000",
                "Userdata HiUser",
                "Groupdata HiGroup",
                "Topicdata HiTopic"
            ],
            Qos(
                Policy.Reliability.Reliable(max_blocking_time=10000000),
                Policy.Durability.TransientLocal,
                Policy.History.KeepLast(depth=10),
                Policy.ResourceLimits(max_samples=100, max_instances=-1, max_samples_per_instance=100),
                Policy.PresentationAccessScope.Topic(coherent_access=True, ordered_access=False),
                Policy.Lifespan(lifespan=1000000),
                Policy.Deadline(deadline=1000000000),
                Policy.LatencyBudget(budget=10000000),
                Policy.Ownership.Exclusive,
                Policy.OwnershipStrength(strength=20),
                Policy.Liveliness.ManualByParticipant(lease_duration=100000),
                Policy.TimeBasedFilter(filter_time=100000),
                Policy.Partition(partitions=('Hello', 'world')),
                Policy.TransportPriority(priority=1),
                Policy.DestinationOrder.BySourceTimestamp,
                Policy.WriterDataLifecycle(autodispose=False),
                Policy.ReaderDataLifecycle(autopurge_nowriter_samples_delay=10000000, autopurge_disposed_samples_delay=10000000),
                Policy.DurabilityService(cleanup_delay=100000, history=Policy.History.KeepLast(depth=100),
                                            max_samples=2000, max_instances=1000, max_samples_per_instance=1000),
                Policy.Userdata(data=b'HiUser'),
                Policy.Groupdata(data=b'HiGroup'),
                Policy.Topicdata(data=b'HiTopic'),
            )
        )
    ]

    for (input, result) in tests:
        pubsub, ddsls = run_pubsub_ddsls(["-T", "test", "-q", ' '.join(input)],
                                         ["-a"],
                                         runtime=2)
        for policy in result:
            assert str(policy) in ddsls["stdout"]

        assert pubsub["status"] == 0


def test_parse_qos_compatible_expressions():
    tests = [
                (
                    [
                        "Reliability.Reliable 10,",
                        "History.KeepLast: 10",
                        "ResourceLimits 100, -1 100",
                        "PresentationAccessScope.Topic 1 no",
                        "liFespAn seconds=1000;days=12",
                        "deadline seconds=1",
                        "WriterDataLifecycle off"
                    ],
                    [
                        Policy.Reliability.Reliable(max_blocking_time=10),
                        Policy.History.KeepLast(depth=10),
                        Policy.ResourceLimits(max_samples=100, max_instances=-1, max_samples_per_instance=100),
                        Policy.PresentationAccessScope.Topic(coherent_access=True, ordered_access=False),
                        Policy.Lifespan(lifespan=1037800000000000),
                        Policy.Deadline(deadline=1000000000),
                        Policy.WriterDataLifecycle(autodispose=False)
                    ]
                )
            ]

    for (input, result) in tests:
        pubsub, ddsls = run_pubsub_ddsls(["-T", "test", "-q", ' '.join(input)],
                                         ["-a"],
                                         runtime=2)
        for policy in result:
            assert str(policy) in ddsls["stdout"]

        assert pubsub["status"] == 0


def test_multiple_qoses():
    pubsub, ddsls = run_pubsub_ddsls(["-T", "test", "--qos", "Durability.TransientLocal", "Userdata HelloWorld"],
                                     ["-a"],
                                     runtime=2)

    assert "Userdata(data=b'HelloWorld')" in ddsls["stdout"]
    assert "Durability.TransientLocal" in ddsls["stdout"]

    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0


def test_qos_special_cases():
    pubsub, ddsls = run_pubsub_ddsls(["-T", "test", "--qos", "PresentationAccessScope.Topic:", "True, False",
                                      "Partition", "test, parti, 33",
                                      "DurabilityService", "seconds=1, history.keeplast 10, 100, 10, 10",
                                      "Topicdata", "helloTopic"],
                                     ["-a"],
                                     runtime=2)

    assert "PresentationAccessScope.Topic(coherent_access=True, ordered_access=False)" in ddsls["stdout"]
    assert "Partition(partitions=('test', 'parti', '33'))" in ddsls["stdout"]
    assert "DurabilityService(cleanup_delay=1000000000, history=Policy.History.KeepLast(depth=10), \
max_samples=100, max_instances=10, max_samples_per_instance=10)" in ddsls["stdout"]
    assert "Topicdata(data=b'helloTopic')" in ddsls["stdout"]

    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0


def test_topic_qos():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "topic",
                                          "--qos", "Durability.TransientLocal"],
                                         ["-t", "dcpspublication"], runtime=2)

    assert "Durability.TransientLocal" in ddsls_pub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "topic",
                                          "--qos", "Durability.TransientLocal"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "Durability.TransientLocal" in ddsls_sub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0


def test_topic_multiple_qoses():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "topic",
                                          "--qos", "Durability.TransientLocal", "ResourceLimits", "100, 10, 10"],
                                         ["-t", "dcpspublication"], runtime=2)

    assert "Durability.TransientLocal" in ddsls_pub["stdout"]
    assert "ResourceLimits(max_samples=100, max_instances=10, max_samples_per_instance=10)" in ddsls_pub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "topic",
                                          "--qos", "Durability.TransientLocal", "ResourceLimits", "100, 10, 10"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "Durability.TransientLocal" in ddsls_sub["stdout"]
    assert "ResourceLimits(max_samples=100, max_instances=10, max_samples_per_instance=10)" in ddsls_sub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0


def test_publisher_qos():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "publisher",
                                          "-q", "PresentationAccessScope.Instance", "False, True"],
                                         ["-t", "dcpspublication"], runtime=2)

    assert "PresentationAccessScope.Instance(coherent_access=False, ordered_access=True)" in ddsls_pub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "publisher",
                                          "-q", "PresentationAccessScope.Instance", "False, True"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "PresentationAccessScope.Instance(coherent_access=False, ordered_access=True)" not in ddsls_sub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0


def test_publisher_multiple_qoses():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "publisher",
                                          "-q", "PresentationAccessScope.Instance", "False, True",
                                          "Groupdata", "TestPublisherQos"],
                                         ["-t", "dcpspublication"], runtime=2)

    assert "PresentationAccessScope.Instance(coherent_access=False, ordered_access=True)" in ddsls_pub["stdout"]
    assert "Groupdata(data=b'TestPublisherQos')" in ddsls_pub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "publisher",
                                          "-q", "PresentationAccessScope.Instance", "False, True",
                                          "Groupdata", "TestPublisherQos"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "PresentationAccessScope.Instance(coherent_access=False, ordered_access=True)" not in ddsls_sub["stdout"]
    assert "Groupdata(data=b'TestPublisherQos')" not in ddsls_sub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0


def test_subscriber_qos():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "subscriber",
                                          "-q", "PresentationAccessScope.Instance", "False, True"],
                                         ["-t", "dcpspublication"], runtime=2)

    assert "PresentationAccessScope.Instance(coherent_access=False, ordered_access=True)" not in ddsls_pub["stdout"]
    assert pubsub["status"] == 0

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "subscriber",
                                          "-q", "PresentationAccessScope.Instance", "False, True"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "PresentationAccessScope.Instance(coherent_access=False, ordered_access=True)" in ddsls_sub["stdout"]
    assert pubsub["status"] == 0


def test_subscriber_multiple_qoses():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "subscriber",
                                          "-q", "PresentationAccessScope.Instance", "False, True",
                                          "Groupdata", "TestSubscriberQos"],
                                         ["-t", "dcpspublication"], runtime=2)

    assert "PresentationAccessScope.Instance(coherent_access=False, ordered_access=True)" not in ddsls_pub["stdout"]
    assert "Groupdata(data=b'TestSubscriberQos')" not in ddsls_pub["stdout"]
    assert pubsub["status"] == 0

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "subscriber",
                                          "-q", "PresentationAccessScope.Instance", "False, True",
                                          "Groupdata", "TestSubscriberQos"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "PresentationAccessScope.Instance(coherent_access=False, ordered_access=True)" in ddsls_sub["stdout"]
    assert "Groupdata(data=b'TestSubscriberQos')" in ddsls_sub["stdout"]
    assert pubsub["status"] == 0


def test_writer_qos():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "datawriter",
                                          "--qos", "Durability.TransientLocal"],
                                         ["-t", "dcpspublication"], runtime=2)

    assert "Durability.TransientLocal" in ddsls_pub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "datawriter",
                                          "--qos", "Durability.TransientLocal"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "Durability.TransientLocal" not in ddsls_sub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0


def test_writer_multiple_qoses():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "datawriter",
                                          "--qos", "Durability.TransientLocal", "TransportPriority", "10"],
                                         ["-t", "dcpspublication"], runtime=2)

    assert "Durability.TransientLocal" in ddsls_pub["stdout"]
    assert "TransportPriority(priority=10)" in ddsls_pub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "datawriter",
                                          "--qos", "Durability.TransientLocal", "TransportPriority", "10"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "Durability.TransientLocal" not in ddsls_sub["stdout"]
    assert "TransportPriority(priority=10)" not in ddsls_sub["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]
    assert pubsub["status"] == 0


def test_reader_qos():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "datareader",
                                          "--qos", "Durability.TransientLocal"],
                                         ["-t", "dcpspublication"], runtime=2)

    assert "Durability.TransientLocal" not in ddsls_pub["stdout"]
    assert pubsub["status"] == 0

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "datareader",
                                          "--qos", "Durability.TransientLocal"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "Durability.TransientLocal" in ddsls_sub["stdout"]
    assert pubsub["status"] == 0


def test_reader_multiple_qoses():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "datareader",
                                          "--qos", "Durability.TransientLocal", "LatencyBudget", "10"],
                                         ["-t", "dcpspublication"], runtime=2)

    assert "Durability.TransientLocal" not in ddsls_pub["stdout"]
    assert "LatencyBudget(budget=10)" not in ddsls_pub["stdout"]
    assert pubsub["status"] == 0

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "datareader",
                                          "--qos", "Durability.TransientLocal", "LatencyBudget", "10"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "Durability.TransientLocal" in ddsls_sub["stdout"]
    assert "LatencyBudget(budget=10)" in ddsls_sub["stdout"]
    assert pubsub["status"] == 0


def test_qos_help():
    pubsub = run_pubsub(["--qoshelp"])
    assert "Available QoS and usage" in pubsub["stdout"]
    assert "--qos Reliability.BestEffort" in pubsub["stdout"]
    assert "--qos Reliability.Reliable [max_blocking_time<integer>]" in pubsub["stdout"]
    assert "--qos Partition [partitions<Sequence[str]>]" in pubsub["stdout"]
    assert "--qos DurabilityService [cleanup_delay<integer>], [History.KeepAll / History.KeepLast [depth<integer>]], \
[max_samples<integer>], [max_instances<integer>], [max_samples_per_instance<integer>]" in pubsub["stdout"]

    assert pubsub["status"] == 0


def test_write_to_file(tmp_path):
    pubsub = run_pubsub(["-T", "test", "--filename", str(tmp_path / "test_pubsub.json"), "--runtime", "0.5"],
                        text=message+"\nhello\n[]")

    time.sleep(0.5)

    with open(tmp_path / "test_pubsub.json") as f:
        data = json.load(f)

    assert "string" == data["sequence 0"]["type"]
    assert "test" == data["sequence 0"]["keyval"]
    assert "integer" == data["sequence 1"]["type"]
    assert 420 == data["sequence 1"]["keyval"]
    assert "int_array" == data["sequence 2"]["type"]
    assert [4, 2, 0] == data["sequence 2"]["keyval"]
    assert "str_array" == data["sequence 3"]["type"]
    assert ['test', 'str', 'array', 'data', 'struct'] == data["sequence 3"]["keyval"]
    assert "int_sequence" == data["sequence 4"]["type"]
    assert [-1, 183] == data["sequence 4"]["keyval"]
    assert "str_sequence" == data["sequence 5"]["type"]
    assert ['test', 'string', 'sequence'] == data["sequence 5"]["keyval"]
    assert "string" == data["sequence 6"]["type"]
    assert "hello" == data["sequence 6"]["keyval"]
    assert "int_sequence" == data["sequence 7"]["type"]
    assert [] == data["sequence 7"]["keyval"]
    assert pubsub["status"] == 0


# test error messages


def test_not_applicable_entity_qos():
    pubsub, ddsls = run_pubsub_ddsls(["-T", "test",
                                      "--qos", "WriterDataLifecycle", "False",
                                      "-eqos", "subscriber"],
                                     ["-a"], runtime=2)

    assert ("InapplicableQosWarning: The Policy.WriterDataLifecycle(autodispose=False) is not applicable for subscriber"
            in pubsub["stderr"])
    assert "WriterDataLifecycle(autodispose=False)" not in ddsls["stdout"]
    assert "Integer(seq=1, keyval=420)" in pubsub["stdout"]


def test_incompatible_qos():
    pubsub, ddsls_pub = run_pubsub_ddsls(["-T", "test", "-eqos", "subscriber",
                                          "--qos", "PresentationAccessScope.Instance", "False, True"],
                                         ["-t", "dcpspublication"], runtime=2)
    assert "PresentationAccessScope.Instance(coherent_access=False, ordered_access=True)" not in ddsls_pub["stdout"]
    assert "The Qos requested for subscription is incompatible with the Qos offered by publication" in pubsub["stderr"]

    pubsub, ddsls_sub = run_pubsub_ddsls(["-T", "test", "-eqos", "subscriber",
                                          "--qos", "PresentationAccessScope.Instance", "False, True"],
                                         ["-t", "dcpssubscription"], runtime=2)

    assert "PresentationAccessScope.Instance(coherent_access=False, ordered_access=True)" in ddsls_sub["stdout"]
    assert "The Qos requested for subscription is incompatible with the Qos offered by publication" in pubsub["stderr"]

    assert "Integer(seq=1, keyval=420)" not in pubsub["stdout"]
    assert "String" not in pubsub["stdout"]
    assert "IntArray" not in pubsub["stdout"]
    assert "StrArray" not in pubsub["stdout"]
    assert "IntSequence" not in pubsub["stdout"]
    assert "StrSequence" not in pubsub["stdout"]


def test_qos_policy_error():
    pubsub = run_pubsub(["-T", "test", "-q", "Hello"])
    assert "No such policy policy.hello" in pubsub["stderr"]


def test_multiple_qos_policy_error():
    pubsub = run_pubsub(["-T", "test", "-q", "Durability.TransientLocal", "Durability.TransientLocal"])
    assert "ValueError: Multiple Qos policies of type Durability" in pubsub["stderr"]


def test_qos_policy_arguments_error():
    pubsub = run_pubsub(["-T", "test", "-q", "History.KeepLast", "hey"])
    assert "ValueError: invalid literal for int()" in pubsub["stderr"]

    pubsub = run_pubsub(["-T", "test", "-q", "WriterDataLifecycle", "sure"])
    assert "Invalid boolean sure" in pubsub["stderr"]


def test_qos_policy_arguments_number_error():
    pubsub = run_pubsub(["-T", "test", "-q", "History.KeepAll", "10"])
    assert "No such policy policy.10" in pubsub["stderr"]

    pubsub = run_pubsub(["-T", "test", "-q", "History.KeepLast"])
    assert "Exception: Unexpected end of arguments" in pubsub["stderr"]

    pubsub = run_pubsub(["-T", "test", "-q", "History.KeepLast", "10", "20"])
    assert "Exception: No such policy policy.20" in pubsub["stderr"]


def test_subqos_error():
    pubsub = run_pubsub(["-T", "test", "-q", "DurabilityService", "100"])
    assert "Exception: Unexpected end of arguments" in pubsub["stderr"]

    pubsub = run_pubsub(["-T", "test", "-q", "DurabilityService", "100, ResourceLimits 1, 10, 10"])
    assert "Exception: DurabilityService takes a History policy" in pubsub["stderr"]


def test_incompatible_qos_value():
    pubsub = run_pubsub(["-T", "test", "-q", "DurabilityService", "10, History.KeepLast 20, 30, 40, 50"])
    assert "Exception: The arguments inputted are considered invalid for cyclonedds" in pubsub["stderr"]


def test_input_data_error_msg():
    new_msg = "[1,'hello',2]"
    pubsub = run_pubsub(["-T", "test"], text=new_msg)
    assert "Exception: TypeError: Element type inconsistent" in pubsub["stderr"]


def test_select_eqos_without_qos_definitions():
    pubsub = run_pubsub(["-T", "test", "-eqos", "all"])
    assert "The following argument is required: -q/--qos" in pubsub["stderr"]

    pubsub = run_pubsub(["-T", "test", "-eqos", "topic"])
    assert "The following argument is required: -q/--qos" in pubsub["stderr"]


def test_file_open_error(tmp_path):
    data = run_ddsls(["--json", "-a", "--filename", f"{tmp_path}/this/path/denfinitely/doesnot/exist/ever"])
    assert f"Exception: Could not open file {tmp_path}/this/path/denfinitely/doesnot/exist/ever" in data["stderr"]


def test_input_syntax_error():
    pubsub = run_pubsub(["-T", "test"], text="[2,3,")
    assert "Input unrecognizable" in pubsub["stderr"]



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import os
import sys
import datetime
import typing
import argparse
import json
from threading import Thread, Event

from .check_entity_qos import QosPerEntity
from .parse_qos import QosParser
from .topic_manager import TopicManager

from cyclonedds.core import WaitSet
from cyclonedds.domain import DomainParticipant
from cyclonedds.util import duration
from cyclonedds.qos import Qos
from dataclasses import fields


def create_parser(args):
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-T", "--topic", type=str, help="The name of the topic to publish/subscribe to")
    group.add_argument("-D", "--dynamic", type=str, help="Dynamically publish/subscribe to a topic")
    parser.add_argument("-i", "--id", type=int, help="Define the domain participant id")
    parser.add_argument("-f", "--filename", type=str, help="Write results to file in JSON format")
    parser.add_argument("-eqos", "--entityqos", choices=["all", "topic", "publisher", "subscriber",
                        "datawriter", "datareader"], default=None, help="""Select the entites to set the qos.
Choose between all entities, topic, publisher, subscriber, datawriter and datareader. (default: all).
Inapplicable qos will be ignored.""")
    parser.add_argument("-q", "--qos", nargs="+",
                        help="Set QoS for entities, check '--qoshelp' for available QoS and usage\n")
    parser.add_argument("-r", "--runtime", type=float, help="Limit the runtime of the tool, in seconds.")
    group.add_argument("--qoshelp", action="store_true", help=qos_help_msg)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)

    args = parser.parse_args(args)

    if args.qoshelp:
        print(qos_help_msg)
        sys.exit(0)
    if args.entityqos and not args.qos:  # Error when selecting entity qos without defining qos policy
        raise Exception("The following argument is required: -q/--qos")
    return args


def qos_help():
    name_map = {
        int: "integer",
        str: "string",
        float: "float",
        bool: "boolean",
        bytes: "bytes"
    }
    qos_help = []
    for name, policy in Qos._policy_mapper.items():
        policy_name = name.replace("Policy.", "")
        _fields = fields(policy)
        if len(_fields) == 0:
            qos_help.append("--qos " f"{policy_name}\n")
        else:
            out = []
            for f in _fields:
                if f.type in name_map:
                    out.append(f"[{f.name}<{name_map[f.type]}>]")
                else:
                    if f.type.__origin__ is typing.Union:
                        out.append("[History.KeepAll / History.KeepLast [depth<integer>]]")
                    elif f.type is typing.Sequence[str]:
                        out.append(f"[{f.name}<Sequence[str]>]")
                    else:
                        out.append(f"[{f.name}<{f.type}>]")
            qos_help.append("--qos " f"{policy_name} {', '.join(out)}\n")
    return qos_help


qos_help_msg = str(f"""e.g.:
    --qos Durability.TransientLocal
    --qos History.KeepLast 10
    --qos ReaderDataLifecycle 10, 20
    --qos Partition [a, b, 123]
    --qos PresentationAccessScope.Instance False, True
    --qos DurabilityService 1000, History.KeepLast 10, 100, 10, 10
    --qos Durability.TransientLocal History.KeepLast 10\n
Available QoS and usage are:\n {' '.join(map(str, qos_help()))}\n""")


class Worker:
    def __init__(self, work_fn):
        self.txt = None
        self.quit_e = Event()
        self.read_e = Event()
        self.work = Thread(target=work_fn, args=(self,))
        self.work.start()

    def is_stopped(self):
        return self.quit_e.is_set()

    def get_input(self):
        if self.txt is not None:
            txt = self.txt
            self.txt = None
            self.read_e.set()
            return txt
        return None

    def put_input(self, txt):
        self.read_e.clear()
        self.txt = txt
        self.read_e.wait()

    def stop(self):
        self.quit_e.set()
        self.work.join()


def make_work_function(manager, waitset, args):
    def work_fn(worker):
        time_start = datetime.datetime.now()
        v = True
        while v and not worker.is_stopped():
            txt = worker.get_input()
            if txt:
                try:  # Integer or list
                    text = eval(txt)
                    manager.write(text)
                except NameError:  # String
                    manager.write(txt.rstrip("\n"))
                except SyntaxError:
                    raise Exception("Input unrecognizable, please check your input.")
            manager.read()
            waitset.wait(duration(microseconds=20))
            if args.runtime:
                v = datetime.datetime.now() < time_start + datetime.timedelta(seconds=args.runtime)

        # Write to file
        if args.filename:
            try:
                with open(args.filename, 'w') as f:
                    json.dump(manager.track_samples, f, indent=4)
                    print(f"\nResults have been written to file {args.filename}\n")
            except OSError:
                raise Exception(f"Could not open file {args.filename}")

        if not v:
            os._exit(0)
    return work_fn


def main(sys_args):
    args = create_parser(sys_args)
    eqos = QosPerEntity(args.entityqos)
    if args.qos:
        qos = QosParser.parse(args.qos)
        eqos.entity_qos(qos, args.entityqos)

    if args.id is None:
        dp = DomainParticipant()
    else:
        dp = DomainParticipant(args.id)

    waitset = WaitSet(dp)
    manager = TopicManager(args, dp, eqos, waitset)
    if args.topic or args.dynamic:
        try:
            worker = Worker(make_work_function(manager, waitset, args))
            while True:
                txt = input("")
                worker.put_input(txt)
        except (KeyboardInterrupt, IOError, ValueError):
            pass
        except EOFError:
            # stdin closed, if runtime is set wait for that to expire, else exit
            if args.runtime:
                worker.work.join()
        finally:
            worker.stop()

    return 0


def command():
    main(sys.argv[1:])
